/*
 *  linux/arch/m68k/kernel/entry.S
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 * 680x0 support by Hamish Macdonald
 *
 */

/*
 * entry.S  contains the system-call and fault low-level handling routines.
 * This also contains the timer-interrupt handler, as well as all interrupts
 * and faults that can result in a task-switch.
 *
 * NOTE: This code handles signal-recognition, which happens every time
 * after a timer-interrupt and after each system call.
 *
 * Stack layout in '_ret_from_exception':
 *	ptrace needs to have all regs on the stack.
 *	if the order here is changed, it needs to be
 *	updated in fork.c:sys_fork, signal.c:do_signal,
 *	ptrace.c and ptrace.h
 *
 *	This allows access to the syscall arguments in registers d1-d5
 *
 *	 0(sp) - d1
 *	 4(sp) - d2
 *	 8(sp) - d3
 *	 C(sp) - d4
 *	10(sp) - d5
 *	14(sp) - d6
 *	18(sp) - d7
 *	1C(sp) - a0
 *	20(sp) - a1
 *	24(sp) - a2
 *	28(sp) - a3
 *	2C(sp) - a4
 *	30(sp) - a5
 *	34(sp) - a6
 *	38(sp) - d0
 *	3C(sp) - usp
 *	40(sp) - orig_d0
 *	44(sp) - stack adjustment
 *	46(sp) - sr
 *	48(sp) - pc
 *	4C(sp) - format & vector
 */

#include <linux/sys.h>

#include <asm/segment.h>

LOFF_D1 	= 0x00
LOFF_D2 	= 0x04
LOFF_D3 	= 0x08
LOFF_D4 	= 0x0C
LOFF_D5 	= 0x10
LOFF_D6 	= 0x14
LOFF_D7 	= 0x18
LOFF_A0 	= 0x1C
LOFF_A1 	= 0x20
LOFF_A2 	= 0x24
LOFF_A3 	= 0x28
LOFF_A4 	= 0x2C
LOFF_A5 	= 0x30
LOFF_A6 	= 0x34
LOFF_D0 	= 0x38
LOFF_USP	= 0x3C
LOFF_ORIG_D0	= 0x40
LOFF_STKADJ	= 0x44
LOFF_SR 	= 0x46
LOFF_PC 	= 0x48
LOFF_FORMAT	= 0x4C


LCF_MASK	= 0x0001

LENOSYS = 38

/*
 * these are offsets into the task-struct
 */
LTS_STATE	=  0
LTS_COUNTER	=  4
LTS_PRIORITY	=  8
LTS_SIGNAL	= 12
LTS_BLOCKED	= 16
LTS_FLAGS	= 20
LTS_ERRNO	= 24

#include <linux/autoconf.h>

/* the following macro is used when enabling interrupts */
#if defined(CONFIG_ATARI) && !defined(CONFIG_AMIGA) && !defined(CONFIG_MAC)
	/* block out HSYNC on the atari */
#define ALLOWINT 0xfbff
#else
	/* portable version */
#define ALLOWINT 0xf8ff
#endif /* machine compilation types */ 

.globl _system_call, _buserr, _trap
.globl _ret_from_exception, _inthandler, _sys_call_table

.text
.align 4
_buserr:
	clrw	sp@-			| stack adjustment
	subql	#8,sp			| make room for usp and orig_d0
	movel	d0,sp@-
	moveml	d1-a6,sp@-
	movec	usp,a0
	movel	a0,sp@(LOFF_USP)
	moveq	#-1,d0
	movel	d0,sp@(LOFF_ORIG_D0)    | a -1 in the ORIG_D0 field
					| signifies that the stack frame
					| is NOT for syscall

	movel	sp,sp@- 		| stack frame pointer argument
	bsrl	_buserr_c
	addql	#4,sp
	jra	_ret_from_exception

.align 4
_trap:
	clrw	sp@-			| stack adjustment
	subql	#8,sp			| make room for usp and orig_d0
	movel	d0,sp@-
	moveml	d1-a6,sp@-
	movec	usp,a0
	movel	a0,sp@(LOFF_USP)
	moveq	#-1,d0
	movel	d0,sp@(LOFF_ORIG_D0)    | a -1 in the ORIG_D0 field
					| signifies that the stack frame
					| is NOT for syscall
	movel	sp,sp@- 		| stack frame pointer argument
	bsrl	_trap_c
	addql	#4,sp
	jra	_ret_from_exception
.align 4
reschedule:
	pea	_ret_from_exception
	jmp	_schedule

.align 4
_system_call:
	clrw	sp@-			| stack adjustment
	subql	#8,sp			| make room for usp and orig_d0
	pea	(-LENOSYS):w		| default return value in D0
	moveml	d1-a6,sp@-		| save registers
	movec	usp,a0
	movel	a0,sp@(LOFF_USP)        | save User Stack Pointer
	movel	d0,sp@(LOFF_ORIG_D0)    | save original D0 (syscall #)
	movel	d0,d2

	| save top of frame
	pea	sp@
	jbsr	_set_esp0
	addql	#4,sp

	cmpl	#NR_syscalls,d2
	jcc	_ret_from_exception
	lea	_sys_call_table,a2
	movel	a2@(d2:l:4),a2
	tstl	a2
	jeq	_ret_from_exception
	andw	#~LCF_MASK,sp@(LOFF_SR) | assume syscall success
	movel	_current,a3
	clrl	a3@(LTS_ERRNO)
	btst	#5,a3@(LTS_FLAGS+3)     | PF_TRACESYS
	bnes	1f
	jbsr	a2@
	movel	d0,sp@(LOFF_D0)         | save the return value
	jpl	2f
	orw	#LCF_MASK,sp@(LOFF_SR)  | set carry to indicate error
2:
	movel	a3@(LTS_ERRNO),d1
	negl	d1
	jeq	_ret_from_exception
	movel	d1,sp@(LOFF_D0)
	orw	#LCF_MASK,sp@(LOFF_SR)  | set carry to indicate error
	jra	_ret_from_exception
1:	jbsr	_syscall_trace
	jbsr	a2@
	movel	d0,sp@(LOFF_D0)         | save the return value
	jpl	2f
	orw	#LCF_MASK,sp@(LOFF_SR)  | set carry to indicate error
2:
	movel	a3@(LTS_ERRNO),d1
	negl	d1
	jeq	2f
	movel	d1,sp@(LOFF_D0)
	orw	#LCF_MASK,sp@(LOFF_SR)  | set carry to indicate error
2:	jbsr	 _syscall_trace

_ret_from_exception:
	btst	#5,sp@(LOFF_SR)		| check if returning to kernel
	bnes	2f			| if so, skip resched, signals
	tstl	_need_resched
	jne	reschedule
	movel	_current,a2
	cmpl	#_task,a2		| task[0] cannot have signals
	jeq	2f
	bclr	#6,a2@(LTS_FLAGS+3)	| check for delayed trace
	jeq	1f
	bclr	#7,sp@(LOFF_SR)		| clear trace bit in SR
	pea	1			| send SIGTRAP
	movel	a2,sp@-
	pea	5
	jbsr	_send_sig
	addql	#8,sp
	addql	#4,sp
1:
	tstl	a2@(LTS_STATE)		| state
	jne	reschedule
	tstl	a2@(LTS_COUNTER)	| counter
	jeq	reschedule
	movel	a2@(LTS_BLOCKED),d0
	movel	d0,d1			| save blocked in d1 for signal handling
	notl	d0
	andl	a2@(LTS_SIGNAL),d0
	jne	Lsignal_return
2:	movel	sp@(LOFF_USP),a0
	movec	a0,usp
	moveml	sp@+,d1-a6
	movel	sp@+,d0 	| restore d0
	addql	#8,sp		| skip the usp and orig_d0
	addw	sp@+,sp 	| add the stack adjustment
	rte
Lsignal_return:
	movel	sp,sp@-
	movel	d1,sp@-
	bsrl	_do_signal
	addql	#8,sp
	bras	2b


/*
** This is the main interrupt handler, responsible for calling process_int()
** or (*mach_process_int)(), if it is a non-autovector int.
*/
_inthandler:
	moveml	a0-a1/d0-d1,sp@-	|  save registers

	addql	#1,_intr_count
	movew	sp@(22),d0              |  put exception # in d0
	andil	#0xfff,d0		|  mask out format nybble

	cmpw	#0x100,d0		|  check if user interrupt
	jge	Luserint

	subiw	#0x60,d0		|  convert from vector offset...
	lsrl	#2,d0			|  ...to IRQ number

	jeq	Lspurious		|  if zero, spurious interrupt
	movec	msp,a0			|  push interrupt stack frame address
	movel	a0,sp@-
	movel	d0,sp@- 		|  put IRQ # on stack
	jbsr	_process_int		|  process the IRQ
	addql	#8,sp			|  pop parameters off stack
	jra 	1f
Luserint:
	movel	_mach_process_int,a1
	tstl	a1			| if no hook installed
	jeq	Lspurious		| ... then spurious interrupt

	subiw	#0x100,d0		| convert from vector offset...
	lsrl	#2,d0			| ...to IRQ number
	movec	msp,a0			|  push interrupt stack frame address
	movel	a0,sp@-
	movel	d0,sp@-			| put IRQ # on stack
	jsr	a1@			| process the IRQ
	addql	#8,sp			| pop parameters off stack

1:


	/* maybe process software interrupts */
	subql	#1,_intr_count
	jne	1f

	movel	_bh_mask,d0
	andl	_bh_active,d0
	jeq	1f

	movew	sr,sp@-
	addql	#1,_intr_count
	andiw	#(ALLOWINT),sr		| allow interrupts
	jbsr	_do_bottom_half
	movew	sp@+,sr
	subql	#1,_intr_count
1:
	/*
	 * check if returning to user mode.
	 * if so, pop stuff off interrupt stack, switch
	 * from interrupt mode to master mode, push
	 * stack frame on kernel stack and jump to ret_from_exception
	 */
	moveb	sp@(22),d0              | get format word from int stack
	andb	#0xf0,d0		| mask off format
	cmpb	#0x10,d0		| check if throwaway frame
	jeq	Lnotstacked		| if, not: interrupts stacked, rte
	moveml	sp@+,a0-a1/d0-d1	| restore registers
	rte
Lnotstacked:
	movec	msp,a0			| get the master stack pointer
	btst	#0x5,a0@		| supervisor bit set in saved SR?
	jeq	Lnotkernel		| if in kernel, rte
	moveml	sp@+,a0-a1/d0-d1	| restore registers
	rte
Lnotkernel:
	/* Avoid building syscall frame if we don't need to.
	   ret_from_syscall will do the same tests again, but most of
	   the time it will return right away and building and
	   restoring the syscall frame takes more time.  */

	tstl	_need_resched
	jne	Lbuildframe
	/* No need to check for delayed trace, cannot happen.  */
	movel	_current,a0
	tstl	a0@(LTS_STATE)
	jne	Lbuildframe
	tstl	a0@(LTS_COUNTER)
	jeq	Lbuildframe
	movel	a0@(LTS_BLOCKED),d0
	notl	d0
	andl	a0@(LTS_SIGNAL),d0
	jne	Lbuildframe

	/* Nothing to do for this process, return now.  */
	moveml	sp@+,a0-a1/d0-d1	| restore registers
	rte

Lbuildframe:
	moveml	sp@+,a0-a1/d0-d1	| restore registers
	addql	#8,sp			| throw away throwaway stack frame
	oriw	#0x1000,sr		| set master bit in SR (sp is now msp)
	andiw	#(ALLOWINT),sr		| allow interrupts
	clrw	sp@-			| stack adjustment
	subql	#8,sp			| make room for usp and orig_d0
	movel	d0,sp@- 		| save registers as for syscall
	moveml	d1-a6,sp@-
	moveq	#-1,d0
	movel	d0,sp@(LOFF_ORIG_D0)    | indicate stack frame not for syscall
	movec	usp,a0
	movel	a0,sp@(LOFF_USP)        | save usp

	| save top of frame
	pea	sp@
	jbsr	_set_esp0
	addql	#4,sp

	jra	_ret_from_exception	| deliver signals, reschedule etc..
Lspurious:
	addql	#1,_num_spurious	| increment the counter of spurious
					| interrupts.
	subql	#1,_intr_count
	moveml	sp@+,a0-a1/d0-d1	| restore registers
	rte				| return from exception

.data
.align 4
_sys_call_table:
	.long _sys_setup		/* 0 */
	.long _sys_exit
	.long _sys_fork
	.long _sys_read
	.long _sys_write
	.long _sys_open			/* 5 */
	.long _sys_close
	.long _sys_waitpid
	.long _sys_creat
	.long _sys_link
	.long _sys_unlink		/* 10 */
	.long _sys_execve
	.long _sys_chdir
	.long _sys_time
	.long _sys_mknod
	.long _sys_chmod		/* 15 */
	.long _sys_chown
	.long _sys_break
	.long _sys_stat
	.long _sys_lseek
	.long _sys_getpid		/* 20 */
	.long _sys_mount
	.long _sys_umount
	.long _sys_setuid
	.long _sys_getuid
	.long _sys_stime		/* 25 */
	.long _sys_ptrace
	.long _sys_alarm
	.long _sys_fstat
	.long _sys_pause
	.long _sys_utime		/* 30 */
	.long _sys_stty
	.long _sys_gtty
	.long _sys_access
	.long _sys_nice
	.long _sys_ftime		/* 35 */
	.long _sys_sync
	.long _sys_kill
	.long _sys_rename
	.long _sys_mkdir
	.long _sys_rmdir		/* 40 */
	.long _sys_dup
	.long _sys_pipe
	.long _sys_times
	.long _sys_prof
	.long _sys_brk			/* 45 */
	.long _sys_setgid
	.long _sys_getgid
	.long _sys_signal
	.long _sys_geteuid
	.long _sys_getegid		/* 50 */
	.long _sys_acct
	.long _sys_phys
	.long _sys_lock
	.long _sys_ioctl
	.long _sys_fcntl		/* 55 */
	.long _sys_mpx
	.long _sys_setpgid
	.long _sys_ulimit
	.long _sys_olduname
	.long _sys_umask		/* 60 */
	.long _sys_chroot
	.long _sys_ustat
	.long _sys_dup2
	.long _sys_getppid
	.long _sys_getpgrp		/* 65 */
	.long _sys_setsid
	.long _sys_sigaction
	.long _sys_sgetmask
	.long _sys_ssetmask
	.long _sys_setreuid		/* 70 */
	.long _sys_setregid
	.long _sys_sigsuspend
	.long _sys_sigpending
	.long _sys_sethostname
	.long _sys_setrlimit		/* 75 */
	.long _sys_getrlimit
	.long _sys_getrusage
	.long _sys_gettimeofday
	.long _sys_settimeofday
	.long _sys_getgroups		/* 80 */
	.long _sys_setgroups
	.long _sys_select
	.long _sys_symlink
	.long _sys_lstat
	.long _sys_readlink		/* 85 */
	.long _sys_uselib
	.long _sys_swapon
	.long _sys_reboot
	.long _sys_readdir
	.long _sys_mmap			/* 90 */
	.long _sys_munmap
	.long _sys_truncate
	.long _sys_ftruncate
	.long _sys_fchmod
	.long _sys_fchown		/* 95 */
	.long _sys_getpriority
	.long _sys_setpriority
	.long _sys_profil
	.long _sys_statfs
	.long _sys_fstatfs		/* 100 */
	.long _sys_ni_syscall /* ioperm for i386 */
	.long _sys_socketcall
	.long _sys_syslog
	.long _sys_setitimer
	.long _sys_getitimer		/* 105 */
	.long _sys_newstat
	.long _sys_newlstat
	.long _sys_newfstat
	.long _sys_uname
	.long _sys_ni_syscall	/* iopl for i386 */	/* 110 */
	.long _sys_vhangup
	.long _sys_idle
	.long _sys_ni_syscall	/* vm86 for i386 */
	.long _sys_wait4
	.long _sys_swapoff		/* 115 */
	.long _sys_sysinfo
	.long _sys_ipc
	.long _sys_fsync
	.long _sys_sigreturn
	.long _sys_clone		/* 120 */
	.long _sys_setdomainname
	.long _sys_newuname
	.long _sys_ni_syscall	/* modify_ldt for i386 */
	.long _sys_adjtimex
	.long _sys_mprotect		/* 125 */
	.long _sys_sigprocmask
	.long _sys_create_module
	.long _sys_init_module
	.long _sys_delete_module
	.long _sys_get_kernel_syms	/* 130 */
	.long _sys_quotactl
	.long _sys_getpgid
	.long _sys_fchdir
	.long _sys_bdflush
	.long _sys_sysfs		/* 135 */
	.long _sys_personality
	.long 0				/* for afs_syscall */
	.long _sys_setfsuid
	.long _sys_setfsgid
	.long _sys_llseek		/* 140 */
	.space (NR_syscalls-140)*4
