/*
linux/arch/m68k/atari/atasound.c

VoxWare compatible Atari TT DMA sound driver for 680x0 Linux

(c) 1995 by Michael Schlueter & Michael Marte

Michael Schlueter (michael@duck.syd.de) did the basic structure of the VFS
interface and the u-law to signed byte conversion.

Michael Marte (marte@informatik.uni-muenchen.de) did the sound queue,
/dev/mixer, /dev/sndstat and complemented the VFS interface. He would like
to thank:
Michael Schlueter for initial ideas and documentation on the MFP and
the DMA sound hardware.
Therapy? for their CD 'Troublegum' which really made me rock.

The author of atari_nosound, atari_mksound and atari_microwire_cmd is
unknown.
(++roman: That's me... :-)

/dev/sndstat is based on code by Hannu Savolainen, the author of the
VoxWare family of drivers.

This file is subject to the terms and conditions of the GNU General Public
License.  See the file README.legal in the main directory of this archive
for more details.

History:
1995/8/25	first release

1995/9/02	++roman: fixed atari_stram_alloc() call, the timer programming and
			several race conditions
*/


#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/mm.h>

#include <asm/atarihw.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/atariints.h>
#include <asm/bootinfo.h>

#include "atasound.h" 
#include <linux/soundcard.h>


/*
stuff from the old atasound.c
*/


static void atari_nosound( unsigned long ignored )
{
	unsigned char	tmp;
	
	/* turn off generator A in mixer control */
	sound_ym.rd_data_reg_sel = 7;
	tmp = sound_ym.rd_data_reg_sel;
	sound_ym.wd_data = tmp | 0x39;
}	
		

void atari_microwire_cmd( int cmd )
{
	tt_microwire.mask = 0x7ff;
	tt_microwire.data = MW_LM1992_ADDR | cmd;

	/* Busy wait for data being completely sent :-( */
	while( tt_microwire.mask != 0x7ff)
		;
}


#define	PC_FREQ		1192180
#define	PSG_FREQ	125000


void atari_mksound( unsigned int count, unsigned int ticks )
/*
 * Generates sound of some count for some number of clock ticks
 * [count = 1193180 / frequency]
 */
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						     atari_nosound };
	unsigned long flags;
	unsigned char tmp;

	save_flags(flags);
	cli();

	if (count == 0x637 && ticks == HZ/8) {
		/* Special case: These values are used by console.c to
		 * generate the console bell. They are catched here and the
		 * sound actually generated is somehow special: it uses the
		 * generator B and an envelope. No timer is needed therefore
		 * and the bell doesn't disturb an other ongoing sound.
		 */

		/* set envelope duration to 492 ms */
		sound_ym.rd_data_reg_sel = 11;
		sound_ym.wd_data = 0;
		sound_ym.rd_data_reg_sel = 12;
		sound_ym.wd_data = 15;
		/* envelope form: max -> min single */
		sound_ym.rd_data_reg_sel = 13;
		sound_ym.wd_data = 9;
		/* set generator B frequency to 2400 Hz */
		sound_ym.rd_data_reg_sel = 2;
		sound_ym.wd_data = 52;
		sound_ym.rd_data_reg_sel = 3;
		sound_ym.wd_data = 0;
		/* set volume of generator B to envelope control */
		sound_ym.rd_data_reg_sel = 9;
		sound_ym.wd_data = 0x10;
		/* enable generator B in the mixer control */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x02) | 0x38;

		restore_flags(flags);
		return;
	}

	del_timer( &sound_timer );

	if (!count) {
		atari_nosound( 0 );
	}
	else {

		/* convert from PC counter value (base frequency 1.193
		   MHz) to PSG period value (base frequency 125 kHz).  */
		int period = (PSG_FREQ * count + PC_FREQ/2) / PC_FREQ;

		if (period > 0xfff) period = 0xfff;

		/* set generator A frequency to 0 */
		sound_ym.rd_data_reg_sel = 0;
		sound_ym.wd_data = period & 0xff;
		sound_ym.rd_data_reg_sel = 1;
		sound_ym.wd_data = (period >> 8) & 0xf;
		/* turn on generator A in mixer control (but not noise
		 * generator!) */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x01) | 0x38;
		/* set generator A level to maximum, no envelope */
		sound_ym.rd_data_reg_sel = 8;
		sound_ym.wd_data = 15;
		
		if (ticks) {
			sound_timer.expires = ticks;
			add_timer( &sound_timer );
		}
	}

	restore_flags(flags);
}


#ifdef CONFIG_ATARI_DMA_SOUND

static int ioctl_return (int *addr, int value);
#define IOCTL_IN(arg)		get_fs_long ((long *)(arg))
#define IOCTL_OUT(arg, ret)	ioctl_return ((int *)arg, ret)


/*
low level sound stuff
*/


struct sound_settings {
	int format;	/* AFMT_{MU_LAW, U8, S8} */
	int speed;	/* DMASND_MODE_{6KHZ, 12KHZ, 25KHZ, 50KHZ} */
	int volume_left, volume_right;	/* [-40, 0] db */
	int bass, treble;	/* [-12, 12] db */
	int stereo;	/* DNASND_MODE_{MONO, STEREO} */
};


struct sound_settings sound;


static void sound_silence (void)
{
	atari_microwire_cmd (MW_LM1992_PSG_HIGH);
		/* mix in PSG signal 1:1 */
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
}


static void sound_init (void)
{
	sound_silence ();
	sound.stereo = DMASND_MODE_MONO;
	sound.speed = DMASND_MODE_6KHZ;
	tt_dmasnd.mode =  DMASND_MODE_MONO | DMASND_MODE_6KHZ;
}


static int sound_set_speed (int speed)
{
	if (speed > 40000) {
		speed = 50066;
		sound.speed = DMASND_MODE_50KHZ;
	} else if (speed > 20000) {
		speed = 25033;
		sound.speed = DMASND_MODE_25KHZ;
	} else if (speed > 7000)  {
		speed = 12517;
		sound.speed = DMASND_MODE_12KHZ;
	} else {
		speed = 6258;
		sound.speed = DMASND_MODE_6KHZ;
	}

	tt_dmasnd.mode = sound.stereo | sound.speed;

	return speed;
}


static int sound_set_format (int format)
{
	if (format != AFMT_MU_LAW && format != AFMT_U8 && format != AFMT_S8)
	  format = AFMT_U8;
	sound.format = format;
	return format;
}



static int sound_set_stereo (int stereo)
{
	if (stereo != 0) stereo = 1;
	sound.stereo = (stereo == 1)? DMASND_MODE_STEREO: DMASND_MODE_MONO;
	tt_dmasnd.mode = sound.stereo | sound.speed;
	return stereo;	
}



#define VOLUME_VOXWARE_TO_DB(v) (((v) < 0)? -40: ((v) > 100)? 0: (((v) * 41 + 50) / 100) - 40)
#define VOLUME_DB_TO_VOXWARE(v) ((((v) + 40) * 100 + 20) / 41)
#define TONE_VOXWARE_TO_DB(v) (((v) < 0)? -12: ((v) > 100)? 12: (((v) + 2) / 4) - 12)
#define TONE_DB_TO_VOXWARE(v) (((v) + 12) * 4)


static int sound_set_volume (int volume)
{
	sound.volume_left = VOLUME_VOXWARE_TO_DB (volume & 0xff);
	atari_microwire_cmd (MW_LM1992_BALLEFT (sound.volume_left));
	sound.volume_right = VOLUME_VOXWARE_TO_DB ((volume & 0xff00) >> 8);
	atari_microwire_cmd (MW_LM1992_BALRIGHT (sound.volume_right));
	return VOLUME_DB_TO_VOXWARE (sound.volume_left) |
		(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8);
}


static int sound_set_treble (int treble)
{
	sound.treble = TONE_VOXWARE_TO_DB (treble);
	atari_microwire_cmd (MW_LM1992_TREBLE (sound.treble));
	return TONE_DB_TO_VOXWARE (sound.treble);
}


static int sound_set_bass (int bass)
{
	sound.bass = TONE_VOXWARE_TO_DB (bass);
	atari_microwire_cmd (MW_LM1992_BASS (sound.bass));
	return TONE_DB_TO_VOXWARE (sound.bass);
}


static unsigned char ulaw_2_dma [] = {
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -128, 
  -128, -128, -128, -128, -128, -128, -128, -126, 
  -123, -119, -115, -111, -107, -103,  -99,  -95, 
   -91,  -87,  -83,  -79,  -75,  -71,  -67,  -63, 
   -60,  -58,  -56,  -54,  -52,  -50,  -48,  -46, 
   -44,  -42,  -40,  -38,  -36,  -34,  -32,  -30, 
   -28,  -27,  -26,  -25,  -24,  -23,  -22,  -21, 
   -20,  -19,  -18,  -17,  -16,  -15,  -14,  -13, 
   -13,  -12,  -12,  -11,  -11,  -10,  -10,   -9, 
    -9,   -8,   -8,   -7,   -7,   -6,   -6,   -5, 
    -5,   -5,   -4,   -4,   -4,   -4,   -3,   -3, 
    -3,   -3,   -2,   -2,   -2,   -2,   -1,   -1, 
    -1,   -1,   -1,   -1,    0,    0,    0,    0, 
     0,    0,    0,    0,    0,    0,    0,    0, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   127,  127,  127,  127,  127,  127,  127,  127, 
   124,  120,  116,  112,  108,  104,  100,   96, 
    92,   88,   84,   80,   76,   72,   68,   64, 
    61,   59,   57,   55,   53,   51,   49,   47, 
    45,   43,   41,   39,   37,   35,   33,   31, 
    29,   28,   27,   26,   25,   24,   23,   22, 
    21,   20,   19,   18,   17,   16,   15,   14, 
    14,   13,   13,   12,   12,   11,   11,   10, 
    10,    9,    9,    8,    8,    7,    7,    6, 
     6,    6,    5,    5,    5,    5,    4,    4, 
     4,    4,    3,    3,    3,    3,    2,    2, 
     2,    2,    2,    2,    1,    1,    1,    1, 
     1,    1,    1,    1,    0,    0,    0,    0, 
};


static void
sound_translate (unsigned char *buff, unsigned long n)
{
	unsigned long i;

	switch (sound.format) {

	case AFMT_MU_LAW:
		n /= 8;
		for (i = 0; i < n; ++i) {
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
			*buff = ulaw_2_dma [*buff]; buff++;
		}
		break;

	case AFMT_U8:
		n /= 8;
		for (i = 0; i < n; ++i) {
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
			*buff -= 128; buff ++;
		}
		break;

	case AFMT_S8: break;

	}
}


/*
/dev/mixer abstraction
*/


struct sound_mixer {
	int busy;
};


struct sound_mixer mixer;


static void mixer_init (void)
{
	mixer.busy = 0;
	atari_microwire_cmd (MW_LM1992_VOLUME (0));
	sound.volume_left = 0;
	atari_microwire_cmd (MW_LM1992_BALLEFT (0));
	sound.volume_right = 0;
	atari_microwire_cmd (MW_LM1992_BALRIGHT (0));
	sound.treble = 0;
	atari_microwire_cmd (MW_LM1992_TREBLE (0));
	sound.bass = 0;
	atari_microwire_cmd (MW_LM1992_BASS (0));
}


static int mixer_open (int open_mode)
{
	if (mixer.busy) return -EBUSY;
	mixer.busy = 1;
	return 0;
}


static int mixer_release (void)
{
	mixer.busy = 0;
	return 0;
}


static int mixer_ioctl (struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch (cmd) {

	case SOUND_MIXER_READ_DEVMASK:
		return IOCTL_OUT (arg, (1 << SOUND_MIXER_VOLUME) | (1 << SOUND_MIXER_TREBLE) | (1 << SOUND_MIXER_BASS));
	case SOUND_MIXER_READ_RECMASK:
		return IOCTL_OUT (arg, 0);
	case SOUND_MIXER_READ_STEREODEVS:
		return IOCTL_OUT (arg, 1 << SOUND_MIXER_VOLUME);

	case SOUND_MIXER_READ_VOLUME:
		return IOCTL_OUT (arg,
			VOLUME_DB_TO_VOXWARE (sound.volume_left) |
			(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8));
	case SOUND_MIXER_READ_BASS:
		return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.bass));
	case SOUND_MIXER_READ_TREBLE:
		return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.treble));

	case SOUND_MIXER_WRITE_VOLUME:
		return IOCTL_OUT (arg, sound_set_volume (IOCTL_IN (arg)));
	case SOUND_MIXER_WRITE_BASS:
		return IOCTL_OUT (arg, sound_set_bass (IOCTL_IN (arg)));
	case SOUND_MIXER_WRITE_TREBLE:
		return IOCTL_OUT (arg, sound_set_treble (IOCTL_IN (arg)));
	
	}

	return -EINVAL;
}




/*
/dev/sndstat
*/


struct sound_state {
	int busy;
	char buf [512];
	int len, ptr;
};


struct sound_state state;


static int state_puts (const char *s)
{
	int l;
	for (l = 0; l < 256 && s [l]; l++);
		/* l = max (256, strlen (s)) */
	if (state.len + l > sizeof (state.buf)) return 0;
	memcpy (&state.buf [state.len], s, l);
	state.len += l;
	return 1;
}


static int state_puti (int i, int radix)
{
	int l;
	static char digit [] = "0123456789abcdef";
	char buf [11];

	l = 0; buf [sizeof (buf) - 1] = 0;

	if (i < 0) {
		state_puts ("-");
		i = -i;
	}

	do {
		buf [sizeof (buf) - 2 - l] = digit [i % radix]; l++;
		i /= radix;
	} while (i);

	return state_puts (&buf [sizeof (buf) - 1 - l]);
}


static void state_init (void)
{
	state.busy = 0;
}


static int state_open (int open_mode)
{
	if (state.busy) return -EBUSY;

	state.len = 0;
	state.ptr = 0;
	state.busy = 1;

	state_puts ("Atari DMA sound driver:\n");

	state_puts ("\tmfp.int_en_a & 0x20 = 0x");
	state_puti (mfp.int_en_a & 0x20, 16); state_puts ("\n");

	state_puts ("\tmfp.int_mk_a & 0x20 = 0x");
	state_puti (mfp.int_mk_a & 0x20, 16); state_puts ("\n");

	state_puts ("\tmfp.tim_ct_a = ");
	state_puti (mfp.tim_ct_a, 10); state_puts ("\n");

	state_puts ("\tmfp.tim_dt_a = ");
	state_puti (mfp.tim_dt_a, 10); state_puts ("\n");

	state_puts ("\tsound.format = 0x");
	state_puti (sound.format, 16); state_puts ("\n");

	state_puts ("\tsound.speed = 0x");
	state_puti (sound.speed, 16); state_puts ("\n");

	state_puts ("\tsound.stereo = 0x");
	state_puti (sound.stereo, 16); state_puts ("\n");

	state_puts ("\tsound.volume_left = ");
	state_puti (sound.volume_left, 10); state_puts (" db\n");

	state_puts ("\tsound.volume_right = ");
	state_puti (sound.volume_right, 10); state_puts (" db\n");

	state_puts ("\tsound.bass = ");
	state_puti (sound.bass, 10); state_puts (" db\n");

	state_puts ("\tsound.treble = ");
	state_puti (sound.treble, 10); state_puts (" db\n");

	return 0;
}


static int state_release (void)
{
	state.busy = 0;
	return 0;
}


static int state_read (char *dest, int count)
{
	int n = state.len - state.ptr;
	if (n > count) n = count;
	if (n <= 0) return 0;
	memcpy_tofs (dest, &state.buf [state.ptr], n);
	state.ptr += n;
	return n;
}


/*
sound queue stuff, the heart of the driver
*/


struct sound_buffer {
	char *start;
	int size;
};


struct sound_queue {
	struct sound_buffer buf;
	int block_size;
	int front, rear, count, max_count;
	int rear_size;
	int playing;
	struct wait_queue *write_queue, *open_queue, *sync_queue;
	int open_mode;
	int busy, syncing;
};


struct sound_queue sq;


#define sq_block_address(i) ((char *)sq.buf.start + (i) * sq.block_size)
#define SIGNAL_RECEIVED (current->signal & ~current->blocked)
#define NON_BLOCKING(open_mode) (open_mode & O_NONBLOCK)
#define ONE_SECOND 100	/* in jiffies (100ths of a second) */
#define NO_TIME_LIMIT 0xffffffff
#define SLEEP(queue, time_limit) \
	current->timeout = jiffies + (time_limit); \
	interruptible_sleep_on (&queue);
#define WAKE_UP(queue) (wake_up_interruptible (&queue))
#define QUEUE_BLOCK_SIZE 1024
#define SOUND_BUFFER_SIZE (100 * QUEUE_BLOCK_SIZE)

#define	DISABLE_IRQ()			\
	do {				\
		mfp.int_mk_a &= ~0x20;	\
	} while( 0 )

#define	ENABLE_IRQ()			\
	do {				\
		mfp.int_mk_a |= 0x20;	\
	} while( 0 )


static void sq_init (const struct sound_buffer *buf, int block_size)
{
	sq.buf.start = buf->start;
	sq.buf.size = buf->size;
	sq.block_size = block_size;
	sq.front = sq.count = 0;
	sq.rear = -1;
	sq.max_count = buf->size / block_size;
	sq.write_queue = sq.open_queue = sq.sync_queue = 0;
	sq.busy = 0;
	sq.syncing = 0;
	sound_init (); sq.playing = 0;
}


static void sq_delete_front (void)
{
	if (sq.count <= 0) return;
	sq.count --;
	sq.front = (sq.front + 1) % sq.max_count;
}


static void sq_play (void)
{
	char *start, *end;

	DISABLE_IRQ();
	if (sq.playing || sq.count <= 0 ||
	    (sq.count == 1 && sq.rear_size < sq.block_size && !sq.syncing)) {
		ENABLE_IRQ();
		return;
	}

	start = sq_block_address (sq.front);
	end = start + ((sq.count == 1)? sq.rear_size: sq.block_size);
	DMASNDSetBase (VTOP (start));
	DMASNDSetEnd (VTOP (end));
		/* Since only an even number of samples per frame can
		   be played, we might loose one byte here. (TO DO) */
	tt_dmasnd.ctrl = DMASND_CTRL_ON ;
	sq.playing = 1;

	ENABLE_IRQ();
}


static void sq_interrupt (void)
{
	if (sq.playing) {
		sq.playing = 0;
		sq_delete_front ();
		WAKE_UP (sq.write_queue);
			/* At least one block of the queue is free
			   now, so wake up a writing process blocked
			   because of a full queue. */
		sq_play ();
		if (!sq.playing) WAKE_UP (sq.sync_queue);
			/* We are not playing after sq_play (), so
			   there is nothing to play any more.  Wake up
			   a process waiting for audio output to
			   drain. */
	}

#if 0
	mfp.int_sv_a = 0xdf;
		/* Clear the interrupt-in-service bit.  This allows
		   for interrupt requests by the MFP to the CPU for
		   pending interrupts. */
#endif
}




static int sq_write (const char *src, int count)
{
	int written = 0, left;
	
	/* ++roman: We can append to the last incomplete frame only if
	   isn't already played.  So sq.count must be more than 1...
	   Also, we must disable the interrupt after discovering an
	   incomplete frame, else it could finish while we append to
	   it.  */
	DISABLE_IRQ();
	if (sq.count > 1) {
		left = sq.block_size - sq.rear_size;
		if (left > count) left = count;
		if (left > 0) {
			char *dest = sq_block_address (sq.rear) + sq.rear_size;
			memcpy_fromfs (dest, src, left);
			sound_translate (dest, left);
			sq.rear_size = sq.block_size;
			count -= left;
			src += left;
			written += left;
		}
	}
	ENABLE_IRQ();
	
	sq_play ();

	while (count > 0) {
		if (sq.count == sq.max_count) {
			if (NON_BLOCKING (sq.open_mode))
				return (written > 0)? written: -EAGAIN;
			SLEEP (sq.write_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)
				return (written > 0)? written: -EINTR;
		} else {
			char *dest;
			DISABLE_IRQ();
			sq.count ++;
			sq.rear = (sq.rear + 1) % sq.max_count;
			left = (sq.block_size > count)? count: sq.block_size;
			dest = sq_block_address (sq.rear);
			sq.rear_size = left;
			/* ++roman: If there are at least 2 frames to
			   play before the one we're going to copy and
			   translate now, it seems safe to enable the
			   interrupt already here, assuming that it
			   won't start to play this frame before we're
			   done here.  */
			if (sq.count > 3) ENABLE_IRQ();
			memcpy_fromfs (dest, src, left);
			sound_translate (dest, left);
			ENABLE_IRQ();
			count -= left;
			src += left;
			written += left;
			sq_play ();
		}
	}

	return written;
}


static int sq_open (int open_mode)
{
	if (sq.busy) {
		if (NON_BLOCKING (open_mode)) return -EBUSY;
		while (sq.busy) {
			SLEEP (sq.open_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)  return -EINTR;
		}
	}
	sq.open_mode = open_mode;
	sq.busy = 1;
	return 0;
}


static void sq_reset (void)
{
	sound_silence ();
	sq.playing = 0;
	while (sq.count > 0) sq_delete_front ();
}


static int sq_sync (void)
{
	int rc = 0;

	sq.syncing = 1;
	sq_play ();

	while (sq.playing) {
		SLEEP (sq.sync_queue, ONE_SECOND);
		if (SIGNAL_RECEIVED) {
			/* While waiting for audio output to drain, an
			   interrupt occured.  Stop audio output
			   immediately and clear the queue.  */
			sq_reset ();
			rc = -EINTR;
			break;
		}
	}

	sq.syncing = 0;
	return rc;
}


static int sq_release (void)
{
	int rc = 0;
	if (sq.busy) {
		rc = sq_sync ();
		sq.busy = 0;
		WAKE_UP (sq.open_queue);
			/* Wake up a process waiting for the queue
			   beeing released.  Note: There may be
			   several processes waiting for a call to
			   open () returning.  */
	}
	return rc;
}


/*
high level sound stuff
*/


static int
sound_open (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: return state_open (file->f_flags);
	case SND_DEV_CTL: return mixer_open (file->f_flags);
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: {
	    	int rc = sq_open (file->f_flags);
	    	if (rc == 0) {
			sound.format = ((dev & 0x0f) == SND_DEV_AUDIO)? AFMT_MU_LAW: AFMT_U8;
			sound_init ();
		}
		return rc;
	}
	default: return -ENXIO;
	}

}


static void
sound_release (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: state_release (); return;
	case SND_DEV_CTL: mixer_release (); return;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: sq_release (); return;
	default: printk ("sound_release: Unknown minor device %d\n", dev);
	}
}


static int
sound_lseek (struct inode *inode, struct file *file, off_t offset, int orig)
{
	return -EPERM;
}


static int
sound_read (struct inode *inode, struct file *file, char *buf, int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: return state_read (buf, count);
	case SND_DEV_CTL:
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return -EPERM;
	default:
		printk ("sound_read: Unknown minor device %d\n", dev);
		return -EINVAL;
	}
}


static int
sound_write (struct inode *inode, struct file *file, char *buf, int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS:
	case SND_DEV_CTL: return -EPERM;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return sq_write(buf, count);
	default:
		printk ("sound_write: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

}


static int
sound_select (struct inode *inode, struct file *file, int sel_type,
	      select_table *wait)
{
	return -EPERM;
}


static int ioctl_return (int *addr, int value)
{
	int error;

	if (value < 0) return value;

	error = verify_area (VERIFY_WRITE, addr, sizeof (int));
	if (error) return error;

	put_fs_long (value, (long*)&((addr)[0]));
	return 0;
}


static int
sound_ioctl (struct inode *inode, struct file *file,
             unsigned int cmd, unsigned long arg)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {

	case SND_DEV_STATUS: return -EPERM;
	case SND_DEV_CTL: return mixer_ioctl (inode, file, cmd, arg);

	case SND_DEV_AUDIO:
	case SND_DEV_DSP:
		switch (cmd) {

		case SNDCTL_DSP_RESET: sq_reset (); return 0;

		case SNDCTL_DSP_POST:
		case SNDCTL_DSP_SYNC: return sq_sync ();

		case SNDCTL_DSP_SPEED:
			return IOCTL_OUT (arg, sound_set_speed (IOCTL_IN (arg)));
		case SNDCTL_DSP_STEREO:
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg)));
		case SOUND_PCM_WRITE_CHANNELS:
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg) - 1) + 1);

		case SNDCTL_DSP_SETFMT:
			return IOCTL_OUT (arg, sound_set_format (IOCTL_IN (arg)));
		case SNDCTL_DSP_GETFMTS:
			return IOCTL_OUT (arg, AFMT_MU_LAW | AFMT_U8 | AFMT_S8);

		case SNDCTL_DSP_GETBLKSIZE:
			return IOCTL_OUT (arg, 10240);
		case SNDCTL_DSP_SUBDIVIDE:
		case SNDCTL_DSP_SETFRAGMENT: break;

		default: return mixer_ioctl (inode, file, cmd, arg);
		}
		break;

	default:
		printk ("sound_ioctl: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

	return( -EINVAL );
}




static struct file_operations sound_fops =
{
	sound_lseek,
	sound_read,
	sound_write,
	NULL,			/* readdir */
	sound_select,
	sound_ioctl,
	NULL,			/* mmap */
	sound_open,
	sound_release,
	NULL			/* fsync */
};


long
soundcard_init (long mem_start)
{
	struct sound_buffer buf;
	
	if (ATARIHW_PRESENT( PCM_8BIT ) &&
	    ATARIHW_PRESENT( MICROWIRE ))
	{
		if ((mfp.int_en_a & mfp.int_mk_a & 0x20) ==0 )
		{
			/* Register driver with the VFS. */
			register_chrdev (SOUND_MAJOR, "sound", &sound_fops);
			printk ("Atari DMA sound driver installed.\n");

			/* Set up sound queue, /dev/audio and
			   /dev/dsp.  TO DO: Allocate buffer space
			   dynamically.  But I do not see how to do
			   this, since kmalloc will not allocate more
			   than 4K and vmalloc does not guarantee
			   memory to be physically contiguous, which
			   is necessary for DMA transfer. */
			/* ++roman: It would be veeery difficult to
			   allocate these buffers dynamically...  Need
			   not only contigous pages, but also in
			   ST-Ram...  */
			buf.start = (char *) atari_stram_alloc (
				SOUND_BUFFER_SIZE, &mem_start);
			buf.size = SOUND_BUFFER_SIZE;
			sq_init (&buf, QUEUE_BLOCK_SIZE);
        
			/* Set up /dev/sndstat. */
			state_init ();

			/* Set up /dev/mixer. */
			mixer_init ();

			/* Set up timer A.  Timer A will receive a
			   signal upon end of playing from the sound
			   hardware.  Furthermore Timer A is able to
			   count events and will cause an interrupt
			   after a programmed number of events.  So
			   all we need to keep the music playing is to
			   provide the sound hardware with new data
			   upon an interrupt from timer A. */
			mfp.tim_ct_a = 0;	/* ++roman: Stop timer before programming! */
			mfp.tim_dt_a = 1;	/* Cause interrupt after first event. */
			mfp.tim_ct_a = 8;	/* Turn on event counting. */
			mach_add_isr (IRQ_MFP_TIMA, (isrfunc)sq_interrupt, 0, NULL);
						/* Register interrupt handler. */
			mfp.int_en_a |= 0x20;	/* Turn interrupt on. */
			mfp.int_mk_a |= 0x20;
		} else
			printk ("Atari DMA sound driver: Timer A interrupt already in use\n");
	}

	return mem_start;
}


void
sound_setup (char *str, int *ints)
{
        /* check the bootstrap parameter for sound= */
	/* at the moment there are no parameters to configure */
}

#endif /* CONFIG_ATARI_DMA_SOUND */
