/*
 * ataints.c -- atari Linux interrupt handling code
 *
 * 5/2/94 Roman Hodek:
 *  Added support for TT interrupts; setup for TT SCU (may someone has
 *  twiddled there and we won't get the right interrupts :-()
 *
 *  Major change: The device-independant code in m68k/ints.c didn't know
 *  about non-autovec ints yet. It hardcoded the number of possible ints to
 *  7 (IRQ1...IRQ7). But the Atari has lots of non-autovec ints! I made the
 *  number of possible ints a constant defined in interrupt.h, which is
 *  47 for the Atari. So we can call add_isr() for all Atari interrupts just
 *  the normal way. Additionally, all vectors >= 48 are initialized to call
 *  trap() instead of inthandler(). This must be changed here, too.
 *
 * 1995-07-16 Lars Brinkhoff <f93labr@dd.chalmers.se>:
 *  Corrected a bug in atari_add_isr() which rejected all SCC
 *  interrupt sources if there were no TT MFP!
 *
 * 12/13/95: New interface functions atari_level_triggered_int() and
 *  atari_register_vme_int() as support for level triggered VME interrupts.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/kernel_stat.h>

#include <asm/system.h>
#include <asm/traps.h>

#include <asm/atarihw.h>
#include <asm/atariints.h>
#include <asm/atari_stdma.h>
#include <asm/bootinfo.h>
#include <asm/irq.h>

/* isrfunc variables for atari interrupt sources */
static isr_node_t *ata_lists[NUM_ATARI_SOURCES];

/* counters for atari interrupts */
unsigned long atastats_interrupts[NUM_ATARI_SOURCES];

/* next free interrupt vector
 * vectors from 0x70 can be allocated by atari_register_vme_int()
 */
static int	next_free_vme_vec = VME_SOURCE_BASE;

/* check for valid int number (complex, sigh...) */
#define	IS_VALID_INTNO(n)										\
	/* standard ints... */										\
	(((n) < VME_SOURCE_BASE &&									\
	  /* for these: TT-MFP int => TT-MFP is present */			\
	  (!((n) >= TTMFP_SOURCE_BASE && (n) < SCC_SOURCE_BASE) ||	\
	   ATARIHW_PRESENT(TT_MFP)) &&								\
	  /* and: SCC int => even number */							\
	  (!((n) >= SCC_SOURCE_BASE) || !((n) & 1)))				\
	 ||															\
	 /* ...or registered VME ints */							\
	 ((n) >= VME_SOURCE_BASE && (n) < next_free_vme_vec))

/*
 * void atari_init_INTS (void)
 *
 * Parameters:	None
 *
 * Returns:	Nothing
 *
 * This function should be called during kernel startup to initialize
 * the atari IRQ handling routines.
 */

/* GK:
 * HBL IRQ handler for Falcon. Nobody needs it :-)
 * ++andreas: raise ipl to disable further interrupts.
 */
asmlinkage void falcon_hblhandler(void);
asm(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(falcon_hblhandler) ":
	moveml	%d0/%a0,%sp@-
	btst	#4,%sp@(14)	/* throwaway frame? */
	jeq	1f
	movec	%msp,%a0		/* yes, use master stack */
	jra	2f
1:	lea	%sp@(8),%a0	/* else, use interrupt stack */
2:	movew	%a0@,%d0
	andw	#0xf8ff,%d0
	orw	#0x0200,%d0	/* set saved ipl to 2 */
	movew	%d0,%a0@
	moveml	%sp@+,%d0/%a0
	rte");

extern void atari_microwire_cmd( int cmd );

void atari_init_INTS (void)

{
	int i;

	/* initialize the isr list */
	for( i = 0; i < NUM_ATARI_SOURCES; ++i ) {
		ata_lists[i] = NULL;
		atastats_interrupts[i] = 0;
	}
	
	/* Initialize the MFP(s) */
	
#ifdef ATARI_USE_SOFTWARE_EOI
	mfp.vec_adr  = 0x48;	/* Software EOI-Mode */
#else
	mfp.vec_adr  = 0x40;	/* Automatic EOI-Mode */
#endif
	mfp.int_en_a =		    /* turn off MFP-Ints */
	mfp.int_en_b = 0x00;
	mfp.int_mk_a =			/* no Masking */
	mfp.int_mk_b = 0xff;

	if (ATARIHW_PRESENT(TT_MFP)) {
#ifdef ATARI_USE_SOFTWARE_EOI
		tt_mfp.vec_adr  = 0x58;		/* Software EOI-Mode */
#else
		tt_mfp.vec_adr  = 0x50;		/* Automatic EOI-Mode */
#endif
		tt_mfp.int_en_a =			/* turn off MFP-Ints */
		tt_mfp.int_en_b = 0x00;
		tt_mfp.int_mk_a =			/* no Masking */
		tt_mfp.int_mk_b = 0xff;
	}
	
	if (ATARIHW_PRESENT(SCC)) {
		scc.cha_a_ctrl = 9;
		asm __volatile__ ("nop; nop; nop; nop; nop; nop");
		scc.cha_a_ctrl = (char) 0xc0; /* hardware reset */
	}

	if (ATARIHW_PRESENT(SCU)) {
		/* init the SCU if present */
		tt_scu.sys_mask = 0x10;		/* enable VBL (for the cursor) and
									 * disable HSYNC interrupts (how
									 * needs them?)  MFP and SCC are
									 * enabled in VME mask
									 */
		tt_scu.vme_mask = 0x60;		/* enable MFP and SCC ints */
	}
	else {
		/* If no SCU, the HSYNC interrupt needs to be disabled this
		 * way. (Else _inthandler in kernel/sys_call.S gets overruns)
		 */
		vectors[VEC_INT2] = falcon_hblhandler;
	}

	if (ATARIHW_PRESENT(PCM_8BIT) && ATARIHW_PRESENT(MICROWIRE)) {
		/* Initialize the LM1992 Sound Controller to enable the PSG sound.
		 * This is misplaced here, it should be in a atasound_init(),
		 * that doesn't exist yet.
		 */
		atari_microwire_cmd( MW_LM1992_PSG_HIGH );
	}
	
	stdma_init();

	/* Initialize the PSG: all sounds off, both ports output */
	sound_ym.rd_data_reg_sel = 7;
	sound_ym.wd_data = 0xff;
}


/*
 * atari_add_isr : add an interrupt service routine for a particular
 *		   machine specific interrupt source.
 *		   If the addition was successful, it returns 1, otherwise
 *		   it returns 0.  It will fail if another routine is already
 *		   bound into the specified source.
 *   Note that the "pri" argument is currently unused.
 */

int atari_add_isr (unsigned long source, isrfunc isr, int pri, void *data,
				   char *name)
{
  unsigned long atari_source = source & ~IRQ_MACHSPEC;
  isr_node_t *p;

  if (!IS_VALID_INTNO(atari_source)) {
      printk ("atari_add_isr: Unknown uservector: %ld\n", atari_source);
      return 0;
  }

  p = new_isr_node();
  p->isr = isr;
  p->pri = pri;
  p->data = data;
  p->name = name;
  p->next = NULL;

  insert_isr (&ata_lists[atari_source], p);

  return 1;
}


/* Bitmask of user interrupt vectors which are generated by the MFP's
   or the SCC. Note: this uses big endian bit numbering.  */
static char atari_edge_triggered_ints[NUM_ATARI_SOURCES/8] =
 /*  ST MFP      TT MFP       SCC    */
 { 0xff, 0xff, 0xff, 0xff, 0xaa, 0xaa, 0, 0 };

void atari_process_int (int level, struct intframe *fp)
{
	ushort sr;
#ifdef ATARI_USE_SOFTWARE_EOI
	int is_edge_int;
	int mask;
#endif

	if (level >= NUM_ATARI_SOURCES)
		panic ("Unknown user interrupt level %d", level);

	asm volatile ("movew %/sr,%0" : "=dm" (sr));
#ifdef ATARI_USE_SOFTWARE_EOI
	asm ("bftst %1{%2,#1}; sne %0"
		 : "=dm" (is_edge_int)
		 : "m" (atari_edge_triggered_ints[0]), "d" (level));
	if (is_edge_int) {
		/* Lower the ipl to let interrupts with higher priority come through */
		asm volatile ("movew %0,%/sr" : : "d" (sr - 0x100));
    }
#endif
	
	kstat.interrupts[(sr >> 8) & 7]++;
	atastats_interrupts[level]++;

	call_isr_list (level, ata_lists[level], fp);

#ifdef ATARI_USE_SOFTWARE_EOI
	if (!is_edge_int)
		return;
	
	/* Raise ipl again to avoid stacked interrupts.  */
	asm volatile ("movew %0,%/sr" : : "dm" (sr));

	/* acknowledge the interrupt */
	mask = ~(1 << (level & 7));
	switch (level >> 3) {
	  case 0:
		mfp.int_sv_b = mask;
		break;
	  case 1:
		mfp.int_sv_a = mask;
		break;
	  case 2:
		tt_mfp.int_sv_b = mask;
		break;
	  case 3:
		tt_mfp.int_sv_a = mask;
		break;
	  case 4:
	  case 5:
		scc.cha_a_ctrl = 070;	/* Reset highest IUS */
		break;
	}
#endif
}


/*
 * atari_level_triggered_int() tells atari_process_int() that one specific
 * interrupt is level-triggered instead of edge-triggered (like the MFP ints.)
 * Lowering the IPL for such an int would cause it to be triggered immediately
 * again, resulting in an endless interrupt loop. Level-triggered ints can
 * come from VME boards.
 */

int atari_int_set_edge_triggered( unsigned long source, int flag )

{
	source &= ~IRQ_MACHSPEC;
	if (!IS_VALID_INTNO(source)) {
		printk( "atari_level_triggered_int: "
				"Unknown uservector: %ld\n", source );
		return( 0 );
    }

	__asm__ __volatile__
		( "bfins %2,%0{%1,#1}"
		  : /* no outpts */
		  : "m" (atari_edge_triggered_ints[0]), "d" (source), "d" (flag) );
	return( 1 );
}


/*
 * atari_register_vme_int() returns the number of a free interrupt vector for
 * hardware with programmable interrupt vector.
 * VME interrupts are by default level-triggered.
 */

unsigned long atari_register_vme_int( void )

{	unsigned long source;
	
	if (next_free_vme_vec == NUM_ATARI_SOURCES)
		return( 0 );

	source = next_free_vme_vec | IRQ_MACHSPEC;
	next_free_vme_vec++;
	return( source );
}


int atari_get_irq_list( char *buf, int len )

{	int			i;
    isr_node_t	*p;

	for( i = 0; i < NUM_ATARI_SOURCES; ++i ) {
		if (!ata_lists[i])
			continue;
		len += sprintf( buf+len, "vec $%02x: %8ld ",
						i+0x40, atastats_interrupts[i] );
		for( p = ata_lists[i]; p; p = p->next ) {
			len += sprintf( buf+len, "%s\n", p->name );
			if (p->next)
				len += sprintf( buf+len, "                  " );
		}
	}
	
	return( len );
}


