/*
linux/arch/m68k/atari/atasound.c

VoxWare compatible Atari TT DMA sound driver for 680x0 Linux

(c) 1995 by Michael Schlueter & Michael Marte

Michael Schlueter (michael@duck.syd.de) did the basic structure of the VFS
interface and the u-law to signed byte conversion.

Michael Marte (marte@informatik.uni-muenchen.de) did the sound queue,
/dev/mixer, /dev/sndstat and complemented the VFS interface. He would like
to thank:
Michael Schlueter for initial ideas and documentation on the MFP and
the DMA sound hardware.
Therapy? for their CD 'Troublegum' which really made me rock.

The author of atari_nosound, atari_mksound and atari_microwire_cmd is
unknown.
(++roman: That's me... :-)

/dev/sndstat is based on code by Hannu Savolainen, the author of the
VoxWare family of drivers.

This file is subject to the terms and conditions of the GNU General Public
License.  See the file README.legal in the main directory of this archive
for more details.

History:
1995/8/25	first release

1995/9/02	++roman: fixed atari_stram_alloc() call, the timer programming
			and several race conditions

1995/9/14	++roman: After some discussion with Michael Schlueter, revised
			the interrupt disabling
			Slightly speeded up U8->S8 translation by using long
			operations where possible
			Added 4:3 interpolation for /dev/audio

1995/9/20	++TeSche: Fixed a bug in sq_write and changed /dev/audio
			converting to play at 12517Hz instead of 6258Hz.

1995/9/23	++TeSche: Changed sq_interrupt() and sq_play() to pre-program
			the DMA for another frame while there's still one
			running. This allows the IRQ response to be
			arbitrarily delayed and playing will still continue.

1995/10/14	++Guenther_Kelleter@ac3.maus.de, ++TeSche: better support for
			Falcon audio (the Falcon doesn't raise an IRQ at the
			end of a frame, but at the beginning instead!). uses
			'if (codec_dma)' in lots of places to simply switch
			between Falcon and TT code.

1995/11/06	++TeSche: started introducing a hardware abstraction scheme
			(may perhaps also serve for Amigas?), can now play
			samples at almost all frequencies by means of a more
			generalized expand routine, takes a good deal of care
			to cut data only at sample sizes, buffer size is now
			a kernel runtime option, implemented fsync() & several
			minor improvements
		++Guenther: useful hints and bugfixes, cross-checked it for
			Falcons
*/


#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/mm.h>

#include <asm/atarihw.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/atariints.h>
#include <asm/bootinfo.h>

#include "atasound.h" 
#include <linux/soundcard.h>



/*
 * stuff from the old atasound.c
 */


static void atari_nosound (unsigned long ignored)
{
	unsigned char	tmp;
	unsigned long flags;
	
	/* turn off generator A in mixer control */
	save_flags(flags);
	cli();
	sound_ym.rd_data_reg_sel = 7;
	tmp = sound_ym.rd_data_reg_sel;
	sound_ym.wd_data = tmp | 0x39;
	restore_flags(flags);
}	
		

void atari_microwire_cmd (int cmd)
{
	tt_microwire.mask = 0x7ff;
	tt_microwire.data = MW_LM1992_ADDR | cmd;

	/* Busy wait for data being completely sent :-( */
	while( tt_microwire.mask != 0x7ff)
		;
}


#define	PC_FREQ		1192180
#define	PSG_FREQ	125000


void atari_mksound (unsigned int count, unsigned int ticks)
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						     atari_nosound };
	/*
	 * Generates sound of some count for some number of clock ticks
	 * [count = 1193180 / frequency]
	 */
	unsigned long flags;
	unsigned char tmp;

	save_flags(flags);
	cli();

	if (count == 0x637 && ticks == HZ/8) {
		/* Special case: These values are used by console.c to
		 * generate the console bell. They are catched here and the
		 * sound actually generated is somehow special: it uses the
		 * generator B and an envelope. No timer is needed therefore
		 * and the bell doesn't disturb an other ongoing sound.
		 */

		/* set envelope duration to 492 ms */
		sound_ym.rd_data_reg_sel = 11;
		sound_ym.wd_data = 0;
		sound_ym.rd_data_reg_sel = 12;
		sound_ym.wd_data = 15;
		/* envelope form: max -> min single */
		sound_ym.rd_data_reg_sel = 13;
		sound_ym.wd_data = 9;
		/* set generator B frequency to 2400 Hz */
		sound_ym.rd_data_reg_sel = 2;
		sound_ym.wd_data = 52;
		sound_ym.rd_data_reg_sel = 3;
		sound_ym.wd_data = 0;
		/* set volume of generator B to envelope control */
		sound_ym.rd_data_reg_sel = 9;
		sound_ym.wd_data = 0x10;
		/* enable generator B in the mixer control */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x02) | 0x38;

		restore_flags(flags);
		return;
	}

	del_timer( &sound_timer );

	if (!count) {
		atari_nosound( 0 );
	}
	else {

		/* convert from PC counter value (base frequency 1.193 MHz)
		 * to PSG period value (base frequency 125 kHz).
		 */
		int period = (PSG_FREQ * count + PC_FREQ/2) / PC_FREQ;

		if (period > 0xfff) period = 0xfff;

		/* set generator A frequency to 0 */
		sound_ym.rd_data_reg_sel = 0;
		sound_ym.wd_data = period & 0xff;
		sound_ym.rd_data_reg_sel = 1;
		sound_ym.wd_data = (period >> 8) & 0xf;
		/* turn on generator A in mixer control (but not noise
		 * generator!) */
		sound_ym.rd_data_reg_sel = 7;
		tmp = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (tmp & ~0x01) | 0x38;
		/* set generator A level to maximum, no envelope */
		sound_ym.rd_data_reg_sel = 8;
		sound_ym.wd_data = 15;
		
		if (ticks) {
			sound_timer.expires = ticks;
			add_timer( &sound_timer );
		}
	}

	restore_flags(flags);
}



#ifdef CONFIG_ATARI_DMA_SOUND



/*** some declarations *******************************************************/


static int ioctl_return (int *addr, int value);
#define IOCTL_IN(arg)		get_fs_long ((long *)(arg))
#define IOCTL_OUT(arg, ret)	ioctl_return ((int *)arg, ret)

#define DMASND_TT	1
#define DMASND_FALCON	2
#define DMASND_AMIGA	3	/* ??? */

#define MAX_CATCH_RADIUS 10
#define MIN_BUFFERS 4
#define MIN_BUFSIZE 4

int catchRadius = 0, numBufs = 4, bufSize = 32;

typedef struct {
	int	type;
	void	(*init)(void);
	void	(*silence)(void);
	int	(*setFormat)(int);
	int	(*setSpeed)(int);
	int	(*setStereo)(int);
	int	(*setVolume)(int);
	int	(*setBass)(int);
	int	(*setTreble)(int);
} MACHINE;

typedef struct {
	int format;	/* AFMT_{MU_LAW | U8 | S8 | U16_BE | S16_BE} */
	int stereo;	/* 0 = mono, 1 = stereo */
	int size;	/* 8/16 bit*/
	int speed;	/* speed */
} SETTINGS;

struct sound_settings {
	MACHINE mach;	/* machine dependend things */
	SETTINGS hard;	/* hardware settings */
	SETTINGS soft;	/* software settings */
	SETTINGS dsp;	/* /dev/dsp default settings */
	void (*copyTranslate)(long, const unsigned char *, long, unsigned char *, long *, long *);
	int volume_left, volume_right;	/* [-40, 0] dB */
	int bass, treble;	/* [-12, 12] dB */
	int minDev;		/* minor device number currently open */
	int bal;		/* balance factor for expanding (not volume!) */
	unsigned long data;	/* data for expanding */
};


struct sound_settings sound;



/*** some low level helpers **************************************************/


static unsigned char ulaw_2_dma [] = {
  -128, -128, -128, -128, -128, -128, -128, -128,
  -128, -128, -128, -128, -128, -128, -128, -128,
  -128, -128, -128, -128, -128, -128, -128, -128,
  -128, -128, -128, -128, -128, -128, -128, -126,
  -123, -119, -115, -111, -107, -103,  -99,  -95,
   -91,  -87,  -83,  -79,  -75,  -71,  -67,  -63,
   -60,  -58,  -56,  -54,  -52,  -50,  -48,  -46,
   -44,  -42,  -40,  -38,  -36,  -34,  -32,  -30,
   -28,  -27,  -26,  -25,  -24,  -23,  -22,  -21,
   -20,  -19,  -18,  -17,  -16,  -15,  -14,  -13,
   -13,  -12,  -12,  -11,  -11,  -10,  -10,   -9,
    -9,   -8,   -8,   -7,   -7,   -6,   -6,   -5,
    -5,   -5,   -4,   -4,   -4,   -4,   -3,   -3,
    -3,   -3,   -2,   -2,   -2,   -2,   -1,   -1,
    -1,   -1,   -1,   -1,    0,    0,    0,    0,
     0,    0,    0,    0,    0,    0,    0,    0,
   127,  127,  127,  127,  127,  127,  127,  127,
   127,  127,  127,  127,  127,  127,  127,  127,
   127,  127,  127,  127,  127,  127,  127,  127,
   127,  127,  127,  127,  127,  127,  127,  127,
   124,  120,  116,  112,  108,  104,  100,   96,
    92,   88,   84,   80,   76,   72,   68,   64,
    61,   59,   57,   55,   53,   51,   49,   47,
    45,   43,   41,   39,   37,   35,   33,   31,
    29,   28,   27,   26,   25,   24,   23,   22,
    21,   20,   19,   18,   17,   16,   15,   14,
    14,   13,   13,   12,   12,   11,   11,   10,
    10,    9,    9,    8,    8,    7,    7,    6,
     6,    6,    5,    5,    5,    5,    4,    4,
     4,    4,    3,    3,    3,    3,    2,    2,
     2,    2,    2,    2,    1,    1,    1,    1,
     1,    1,    1,    1,    0,    0,    0,    0
};


/* ++TeSche: radically changed for new expanding purposes...
 *
 * These two routines now deal with copying/expanding/translating the samples
 * from user space into our buffer at the right frequency. They take care about
 * how much data there's actually to read, how much buffer space there is and
 * to convert samples into the right frequency/encoding. They will only work on
 * complete samples so it may happen they leave some bytes in the input stream
 * if the user didn't write a multiple of the current sample size. They both
 * return the number of bytes they've used from both streams so you may detect
 * such a situation. Luckily all programs should be able to cope with that.
 *
 * I think I've optimized anything as far as one can do in plain C, all
 * variables should fit in registers and the loops are really short. There's
 * one loop for every possible situation. Writing a more generalized and thus
 * parametrized loop would only produce slower code. Feel free to optimize
 * this in assembler if you like. :)
 *
 * I think these routines belong here because they're not yet really hardware
 * independend, especially the fact that the Falcon can play 16bit samples
 * only in stereo is hardcoded in both of them!
 */

static void soundCopyTranslateNormal(long userCount,
				     const unsigned char *userPtr,
				     long bufferCount,
				     unsigned char *bufferPtr,
				     long *userUsed,
				     long *bufferUsed)
{
	long	count;
	unsigned long	data;

	*bufferUsed = bufferCount;
	*userUsed = userCount;

	if (!sound.soft.stereo) switch (sound.soft.format) {

		case AFMT_MU_LAW:
			count = (userCount < bufferCount) ?
				userCount : bufferCount;
			bufferCount -= count;
			userCount -= count;
			while (42) {
				if (count < 1)
					goto done;
				*bufferPtr++ = ulaw_2_dma[get_fs_byte(userPtr++)];
				count--;
			}
			break;

		case AFMT_S8:
			count = (userCount < bufferCount) ?
				userCount : bufferCount;
			memcpy_fromfs(bufferPtr, userPtr, count);
			bufferCount -= count;
			userCount -= count;
			break;

		case AFMT_U8:
			count = (userCount < bufferCount) ?
				userCount : bufferCount;
			bufferCount -= count;
			userCount -= count;
			while (42) {
				if (count < 1)
					goto done;
				*bufferPtr++ = get_fs_byte(userPtr++) ^ 0x80;
				count--;
			}
			break;

		case AFMT_S16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (userCount < 2)
					goto done;
				data = get_fs_word(((unsigned short *)userPtr)++);
				userCount -= 2;
				*((unsigned short *)bufferPtr)++ = data;
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 4;
			}

		case AFMT_U16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (userCount < 2)
					goto done;
				data = get_fs_word(((unsigned short *)userPtr)++) ^ 0x8000;
				userCount -= 2;
				*((unsigned short *)bufferPtr)++ = data;
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 4;
			}

	} else switch (sound.soft.format) {

		case AFMT_MU_LAW:
			count = ((userCount < bufferCount) ?
				userCount : bufferCount) & ~1;
			bufferCount -= count;
			userCount -= count;
			while (42) {
				if (count < 1)
					goto done;
				*bufferPtr++ = ulaw_2_dma[get_fs_byte(userPtr++)];
				count--;
			}

		case AFMT_S8:
			count = ((userCount < bufferCount) ?
				userCount : bufferCount) & ~1;
			memcpy_fromfs(bufferPtr, userPtr, count);
			bufferCount -= count;
			userCount -= count;
			break;

		case AFMT_U8:
			count = ((userCount < bufferCount) ?
				userCount : bufferCount) & ~1;
			bufferCount -= count;
			userCount -= count;
			while (42) {
				if (count < 2)
					goto done;
				*((unsigned short *)bufferPtr)++ = get_fs_word(((unsigned short *)bufferPtr)++) ^ 0x8080;
				count -= 2;
			}

		case AFMT_S16_BE:
			count = ((userCount < bufferCount) ?
				userCount : bufferCount) & ~3;
			memcpy_fromfs(bufferPtr, userPtr, count);
			bufferCount -= count;
			userCount -= count;
			break;

		case AFMT_U16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (userCount < 4)
					goto done;
				*((unsigned long *)bufferPtr)++ = get_fs_long(((unsigned int *)userPtr)++) ^ 0x80008000;
				userCount -= 4;
				bufferCount -= 4;
			}
	}

done:
	*bufferUsed -= bufferCount;
	*userUsed -= userCount;
}


static void soundCopyTranslateExpanding(long userCount,
					const unsigned char *userPtr,
					long bufferCount,
					unsigned char *bufferPtr,
					long *userUsed,
					long *bufferUsed)
{
	/* this should help gcc to stuff everything into registers */
	unsigned long	data = sound.data;
	long		bal = sound.bal;
	long		hSpeed = sound.hard.speed, sSpeed = sound.soft.speed;

	*bufferUsed = bufferCount;
	*userUsed = userCount;

	if (!sound.soft.stereo) switch (sound.soft.format) {

		case AFMT_MU_LAW:
			while (42) {
				if (!bufferCount)
					goto done;
				if (bal < 0) {
					if (!userCount)
						goto done;
					data = ulaw_2_dma[get_fs_byte(userPtr++)];
					userCount--;
					bal += hSpeed;
				}
				*bufferPtr++ = data;
				bufferCount--;
				bal -= sSpeed;
			}

		case AFMT_S8:
			while (42) {
				if (!bufferCount)
					goto done;
				if (bal < 0) {
					if (!userCount)
						goto done;
					data = get_fs_byte(userPtr++);
					userCount--;
					bal += hSpeed;
				}
				*bufferPtr++ = data;
				bufferCount--;
				bal -= sSpeed;
			}

		case AFMT_U8:
			while (42) {
				if (!bufferCount)
					goto done;
				if (bal < 0) {
					if (!userCount)
						goto done;
					data = get_fs_byte(userPtr++) ^ 0x80;
					userCount--;
					bal += hSpeed;
				}
				*bufferPtr++ = data;
				bufferCount--;
				bal -= sSpeed;
			}

		case AFMT_S16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (bal < 0) {
					if (userCount < 2)
						goto done;
					data = get_fs_word(((unsigned short *)userPtr)++);
					userCount -= 2;
					bal += hSpeed;
				}
				*((unsigned short *)bufferPtr)++ = data;
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 4;
				bal -= sSpeed;
			}

		case AFMT_U16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (bal < 0) {
					if (userCount < 2)
						goto done;
					data = get_fs_word(((unsigned char *)userPtr)++) ^ 0x8000;
					userCount -= 2;
					bal += hSpeed;
				}
				*((unsigned short *)bufferPtr)++ = data;
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 4;
				bal -= sSpeed;
			}

	} else switch (sound.soft.format) {

		case AFMT_MU_LAW:
			while (42) {
				if (bufferCount < 2)
					goto done;
				if (bal < 0) {
					if (userCount < 2)
						goto done;
					data = ulaw_2_dma[get_fs_byte(userPtr++)] << 8;
					data |= ulaw_2_dma[get_fs_byte(userPtr++)];
					userCount -= 2;
					bal += hSpeed;
				}
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 2;
				bal -= sSpeed;
			}

		case AFMT_S8:
			while (42) {
				if (bufferCount < 2)
					goto done;
				if (bal < 0) {
					if (userCount < 2)
						goto done;
					data = get_fs_word(((unsigned short *)userPtr)++);
					userCount -= 2;
					bal += hSpeed;
				}
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 2;
				bal -= sSpeed;
			}

		case AFMT_U8:
			while (42) {
				if (bufferCount < 2)
					goto done;
				if (bal < 0) {
					if (userCount < 2)
						goto done;
					data = get_fs_word(((unsigned short *)userPtr)++) ^ 0x8080;
					userCount -= 2;
					bal += hSpeed;
				}
				*((unsigned short *)bufferPtr)++ = data;
				bufferCount -= 2;
				bal -= sSpeed;
			}

		case AFMT_S16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (bal < 0) {
					if (userCount < 4)
						goto done;
					data = get_fs_long(((unsigned int *)userPtr)++);
					userCount -= 4;
					bal += hSpeed;
				}
				*((unsigned long *)bufferPtr)++ = data;
				bufferCount -= 4;
				bal -= sSpeed;
			}

		case AFMT_U16_BE:
			while (42) {
				if (bufferCount < 4)
					goto done;
				if (bal < 0) {
					if (userCount < 4)
						goto done;
					data = get_fs_long(((unsigned int *)userPtr)++) ^ 0x80008000;
					userCount -= 4;
					bal += hSpeed;
				}
				*((unsigned long *)bufferPtr)++ = data;
				bufferCount -= 4;
				bal -= sSpeed;
			}
	}

done:
	sound.bal = bal;
	sound.data = data;
	*bufferUsed -= bufferCount;
	*userUsed -= userCount;
}



/*** low level stuff *********************************************************/


/*
 * TT
 */


static void TTSilence (void)
{
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
	atari_microwire_cmd (MW_LM1992_PSG_HIGH); /* mix in PSG signal 1:1 */
}


static void TTInit (void)
{
	int mode, i, idx;
	int freq[8] = {50066, 25033, 12517, 6258};

	/* search a frequency that fits into the allowed error range */

	idx = -1;
	for (i=0; i<4; i++)
		/* this isn't as much useful for a TT than for a Falcon, but
		 * then it doesn't hurt very much to implement it for a TT too.
		 */
		if ((100 * abs(sound.soft.speed - freq[i]) / freq[i]) < catchRadius)
			idx = i;
	if (idx > -1) {
		sound.soft.speed = freq[idx];
		sound.copyTranslate = soundCopyTranslateNormal;
	} else
		sound.copyTranslate = soundCopyTranslateExpanding;

	TTSilence ();
	sound.hard = sound.soft;

	sound.copyTranslate = soundCopyTranslateExpanding;

	if (sound.hard.speed > 50066) {
		/* we would need to squeeze the sound, but we won't do that */
		sound.hard.speed = 50066;
		mode = DMASND_MODE_50KHZ;
		sound.copyTranslate = soundCopyTranslateNormal;
	} else if (sound.hard.speed > 25033) {
		sound.hard.speed = 50066;
		mode = DMASND_MODE_50KHZ;
	} else if (sound.hard.speed > 12517) {
		sound.hard.speed = 25033;
		mode = DMASND_MODE_25KHZ;
	} else if (sound.hard.speed > 6258) {
		sound.hard.speed = 12517;
		mode = DMASND_MODE_12KHZ;
	} else {
		sound.hard.speed = 6258;
		mode = DMASND_MODE_6KHZ;
	}

	tt_dmasnd.mode = (sound.hard.stereo ?
				DMASND_MODE_STEREO : DMASND_MODE_MONO) |
			  DMASND_MODE_8BIT |
			  mode;

	sound.bal = -sound.soft.speed;
}


static int TTSetFormat (int format)
{
	/* TT sound DMA supports only 8bit modes */

	switch (format) {
		case AFMT_QUERY:
			return sound.soft.format;
		case AFMT_MU_LAW:
		case AFMT_S8:
		case AFMT_U8:
			break;
		default:
			format = AFMT_S8;
	}

	sound.soft.format = format;
	sound.soft.size = 8;
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.format = format;
		sound.dsp.size = 8;
	}
	TTInit();

	return format;
}


#define VOLUME_VOXWARE_TO_DB(v) \
	(((v) < 0)? -40: ((v) > 100)? 0: ((v) * 2) / 5 - 40)
#define VOLUME_DB_TO_VOXWARE(v) ((((v) + 40) * 5 + 1) / 2)


static int TTSetVolume (int volume)
{
	sound.volume_left = VOLUME_VOXWARE_TO_DB (volume & 0xff);
	atari_microwire_cmd (MW_LM1992_BALLEFT (sound.volume_left));
	sound.volume_right = VOLUME_VOXWARE_TO_DB ((volume & 0xff00) >> 8);
	atari_microwire_cmd (MW_LM1992_BALRIGHT (sound.volume_right));
	return VOLUME_DB_TO_VOXWARE (sound.volume_left) |
		(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8);
}


#define TONE_VOXWARE_TO_DB(v) \
	(((v) < 0)? -12: ((v) > 100)? 12: ((v) - 50) * 6 / 25)
#define TONE_DB_TO_VOXWARE(v) (((v) * 25 + ((v) > 0 ? 5 : -5)) / 6 + 50)


static int TTSetBass (int bass)
{
	sound.bass = TONE_VOXWARE_TO_DB (bass);
	atari_microwire_cmd (MW_LM1992_BASS (sound.bass));
	return TONE_DB_TO_VOXWARE (sound.bass);
}


static int TTSetTreble (int treble)
{
	sound.treble = TONE_VOXWARE_TO_DB (treble);
	atari_microwire_cmd (MW_LM1992_TREBLE (sound.treble));
	return TONE_DB_TO_VOXWARE (sound.treble);
}



/*
 * Falcon
 */


static void FalconSilence (void)
{
	/* stop playback, set sample rate 50kHz for PSG sound */
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
	tt_dmasnd.mode = DMASND_MODE_50KHZ | DMASND_MODE_STEREO | DMASND_MODE_8BIT;
	tt_dmasnd.int_div = 0; /* STE compatible divider */
	tt_dmasnd.int_ctrl = 0x0;
	tt_dmasnd.cbar_src = 0x0000; /* no matrix inputs */
	tt_dmasnd.cbar_dst = 0x0000; /* no matrix outputs */
	tt_dmasnd.dac_src = 1; /* connect ADC to DAC, disconnect matrix */
	tt_dmasnd.adc_src = 3; /* ADC Input = PSG */
}


static void FalconInit (void)
{
	int divider, i, idx;
	int freq[8] = {49170, 32780, 24585, 19668, 16390, 12292, 9834, 8195};

	/* search a frequency that fits into the allowed error range */

	idx = -1;
	for (i=0; i<8; i++)
		/* if we will tolerate 3% error 8000Hz->8195Hz (2.38%) would
		 * be playable without expanding, but that now a kernel runtime
		 * option
		 */
		if ((100 * abs(sound.soft.speed - freq[i]) / freq[i]) < catchRadius)
			idx = i;
	if (idx > -1) {
		sound.soft.speed = freq[idx];
		sound.copyTranslate = soundCopyTranslateNormal;
	} else
		sound.copyTranslate = soundCopyTranslateExpanding;

	FalconSilence ();
	sound.hard = sound.soft;

	if (sound.hard.size == 16) {
		/* the Falcon can play 16bit samples only in stereo */
		sound.hard.stereo = 1;
	}

	sound.copyTranslate = soundCopyTranslateExpanding;

	if (sound.hard.speed > 49170) {
		/* we would need to squeeze the sound, but we won't do that */
		sound.hard.speed = 49170;
		divider = 1;
		sound.copyTranslate = soundCopyTranslateNormal;
	} else if (sound.hard.speed > 32780) {
		sound.hard.speed = 49170;
		divider = 1;
	} else if (sound.hard.speed > 24585) {
		sound.hard.speed = 32780;
		divider = 2;
	} else if (sound.hard.speed > 19668) {
		sound.hard.speed = 24585;
		divider = 3;
	} else if (sound.hard.speed > 16390) {
		sound.hard.speed = 19668;
		divider = 4;
	} else if (sound.hard.speed > 12292) {
		sound.hard.speed = 16390;
		divider = 5;
	} else if (sound.hard.speed > 9834) {
		sound.hard.speed = 12292;
		divider = 7;
	} else if (sound.hard.speed > 8195) {
		sound.hard.speed = 9834;
		divider = 9;
	} else {
		sound.hard.speed = 8195;
		divider = 11;
	}
	tt_dmasnd.int_div = divider;

	/* Setup Falcon sound DMA for playback */
	tt_dmasnd.int_ctrl = 0x4; /* Timer A int at play end */
	tt_dmasnd.track_select = 0x0; /* play 1 track, track 1 */
	tt_dmasnd.cbar_src = 0x0001; /* DMA(25MHz) --> DAC */
	tt_dmasnd.cbar_dst = 0x0000;
	tt_dmasnd.rec_track_select = 0;
	tt_dmasnd.dac_src = 2; /* connect matrix to DAC */
	tt_dmasnd.adc_src = 0; /* ADC Input = Mic */

	tt_dmasnd.mode =  (sound.hard.stereo ?
				DMASND_MODE_STEREO : DMASND_MODE_MONO) |
			  ((sound.hard.size == 8) ?
				DMASND_MODE_8BIT : DMASND_MODE_16BIT) |
			  DMASND_MODE_6KHZ;

	sound.bal = -sound.soft.speed;
}


static int FalconSetFormat (int format)
{
	/* Falcon sound DMA supports 8bit and 16bit modees */

	switch (format) {
		case AFMT_QUERY:
			return sound.soft.format;
		case AFMT_MU_LAW:
		case AFMT_U8:
		case AFMT_S8:
			sound.soft.size = 8;
			break;
		case AFMT_S16_BE:
		case AFMT_U16_BE:
			sound.soft.size = 16;
			break;
		default: /* :-) */
			sound.soft.size = 8;
			format = AFMT_S8;
	}

	sound.soft.format = format;
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.format = format;
		sound.dsp.size = sound.soft.size;
	}

	FalconInit();

	return format;
}


/* This is for the Falcon output *attenuation* in 1.5dB steps,
 * i.e. output level from 0 to -22.5dB in -1.5dB steps.
 */
#define VOLUME_VOXWARE_TO_ATT(v) \
	((v) < 0 ? 15 : (v) > 100 ? 0 : 15 - (v) * 3 / 20)
#define VOLUME_ATT_TO_VOXWARE(v) (100 - (v) * 20 / 3)


static int FalconSetVolume (int volume)
{
	sound.volume_left = VOLUME_VOXWARE_TO_ATT(volume & 0xff);
	sound.volume_right = VOLUME_VOXWARE_TO_ATT((volume & 0xff00) >> 8);
	tt_dmasnd.output_atten = sound.volume_left << 8 | sound.volume_right << 4;
	return VOLUME_ATT_TO_VOXWARE(sound.volume_left) |
		VOLUME_ATT_TO_VOXWARE(sound.volume_right) << 8;
}



/*** machine definitions *****************************************************/


MACHINE	machTT =	{DMASND_TT, TTInit, TTSilence, TTSetFormat, NULL,
			NULL, TTSetVolume, TTSetBass, TTSetTreble};
MACHINE	machFalcon =	{DMASND_FALCON, FalconInit, FalconSilence,
			FalconSetFormat, NULL, NULL, FalconSetVolume, NULL,
			NULL};
MACHINE machAmiga =	{DMASND_AMIGA, NULL, NULL, NULL, NULL, NULL, NULL,
			NULL, NULL};



/*** mid level stuff *********************************************************/


static void sound_silence (void)
{
	/* update hardware settings one more */
	(*sound.mach.init)();

	(*sound.mach.silence)();
}


static void sound_init (void)
{
	(*sound.mach.init)();
}


static int sound_set_format (int format)
{
	return (*sound.mach.setFormat)(format);
}


static int sound_set_speed (int speed)
{
	if (speed < 0) {
		return sound.soft.speed;
	}

	sound.soft.speed = speed;
	(*sound.mach.init)();
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.speed = sound.soft.speed;
	}

	return sound.soft.speed;
}


static int sound_set_stereo (int stereo)
{
	if (stereo < 0) {
		return sound.soft.stereo;
	}

	stereo = !!stereo;	/* should be 0 or 1 now */

	sound.soft.stereo = stereo;
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.stereo = stereo;
	}
	(*sound.mach.init)();

	return stereo;
}


static int sound_set_volume (int volume)
{
	return (*sound.mach.setVolume)(volume);
}


static int sound_set_treble (int treble)
{
	sound.treble = TONE_VOXWARE_TO_DB (treble);
	atari_microwire_cmd (MW_LM1992_TREBLE (sound.treble));
	return TONE_DB_TO_VOXWARE (sound.treble);
}


static int sound_set_bass (int bass)
{
	sound.bass = TONE_VOXWARE_TO_DB (bass);
	atari_microwire_cmd (MW_LM1992_BASS (sound.bass));
	return TONE_DB_TO_VOXWARE (sound.bass);
}



/*
 * /dev/mixer abstraction
 */


#define RECLEVEL_VOXWARE_TO_GAIN(v) \
	((v) < 0 ? 0 : (v) > 100 ? 15 : (v) * 3 / 20)
#define RECLEVEL_GAIN_TO_VOXWARE(v) (((v) * 20 + 2) / 3)



struct sound_mixer {
	int busy;
};


static struct sound_mixer mixer;


static void mixer_init (void)
{
	mixer.busy = 0;
	sound.treble = 0;
	sound.bass = 0;
	switch (sound.mach.type) {
		case DMASND_TT:
			atari_microwire_cmd (MW_LM1992_VOLUME (0));
			sound.volume_left = 0;
			atari_microwire_cmd (MW_LM1992_BALLEFT (0));
			sound.volume_right = 0;
			atari_microwire_cmd (MW_LM1992_BALRIGHT (0));
			atari_microwire_cmd (MW_LM1992_TREBLE (0));
			atari_microwire_cmd (MW_LM1992_BASS (0));
			break;
		case DMASND_FALCON:
			sound.volume_left = (tt_dmasnd.output_atten & 0xf00) >> 8;
			sound.volume_right = (tt_dmasnd.output_atten & 0xf0) >> 4;
		case DMASND_AMIGA:
	}
}


static int mixer_open (int open_mode)
{
	if (mixer.busy)
		return -EBUSY;
	mixer.busy = 1;
	return 0;
}


static int mixer_release (void)
{
	mixer.busy = 0;
	return 0;
}


static int mixer_ioctl (struct inode *inode,
			struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch (sound.mach.type) {
		case DMASND_FALCON:
			switch (cmd) {
				case SOUND_MIXER_READ_DEVMASK:
					return IOCTL_OUT (arg, SOUND_MASK_VOLUME | SOUND_MASK_MIC | SOUND_MASK_SPEAKER);
				case SOUND_MIXER_READ_RECMASK:
					return IOCTL_OUT (arg, SOUND_MASK_MIC);
				case SOUND_MIXER_READ_STEREODEVS:
					return IOCTL_OUT (arg, SOUND_MASK_VOLUME | SOUND_MASK_MIC);
				case SOUND_MIXER_READ_CAPS:
					return IOCTL_OUT (arg, SOUND_CAP_EXCL_INPUT);
				case SOUND_MIXER_READ_VOLUME:
					return IOCTL_OUT (arg,
						VOLUME_ATT_TO_VOXWARE(sound.volume_left) |
						VOLUME_ATT_TO_VOXWARE(sound.volume_right) << 8);
				case SOUND_MIXER_READ_BASS:
					return -EINVAL;
				case SOUND_MIXER_READ_TREBLE:
					return -EINVAL;
				case SOUND_MIXER_WRITE_MIC:
					tt_dmasnd.input_gain =
						RECLEVEL_VOXWARE_TO_GAIN(IOCTL_IN(arg) & 0xff) << 4 |
						RECLEVEL_VOXWARE_TO_GAIN(IOCTL_IN(arg) >> 8 & 0xff);
					/* fall thru, return set value */
				case SOUND_MIXER_READ_MIC:
					return IOCTL_OUT (arg,
						RECLEVEL_GAIN_TO_VOXWARE(tt_dmasnd.input_gain >> 4 & 0xf) |
						RECLEVEL_GAIN_TO_VOXWARE(tt_dmasnd.input_gain & 0xf) << 8);
				case SOUND_MIXER_READ_SPEAKER: { int porta;
					cli();
					sound_ym.rd_data_reg_sel = 14;
					porta = sound_ym.rd_data_reg_sel;
					sti();
					return IOCTL_OUT (arg, porta & 0x40 ? 0 : 100);
					}
				case SOUND_MIXER_WRITE_VOLUME:
					return IOCTL_OUT (arg, sound_set_volume (IOCTL_IN (arg)));
				case SOUND_MIXER_WRITE_BASS:
					return -EINVAL;
				case SOUND_MIXER_WRITE_TREBLE:
					return -EINVAL;
				case SOUND_MIXER_WRITE_SPEAKER: { int porta;
					cli();
					sound_ym.rd_data_reg_sel = 14;
					porta = (sound_ym.rd_data_reg_sel & ~0x40) |
					        (IOCTL_IN(arg) < 50 ? 0x40 : 0);
					sound_ym.wd_data = porta;
					sti();
					return IOCTL_OUT (arg, porta & 0x40 ? 0 : 100);
					}
			}
			break;

		case DMASND_TT:
			switch (cmd) {
				case SOUND_MIXER_READ_DEVMASK:
					return IOCTL_OUT (arg,
						SOUND_MASK_VOLUME | SOUND_MASK_TREBLE | SOUND_MASK_BASS |
						((boot_info.bi_atari.mch_cookie >> 16) == ATARI_MCH_TT ?
							SOUND_MASK_SPEAKER : 0));
				case SOUND_MIXER_READ_RECMASK:
					return IOCTL_OUT (arg, 0);
				case SOUND_MIXER_READ_STEREODEVS:
					return IOCTL_OUT (arg, SOUND_MASK_VOLUME);
				case SOUND_MIXER_READ_VOLUME:
					return IOCTL_OUT (arg,
						VOLUME_DB_TO_VOXWARE (sound.volume_left) |
						(VOLUME_DB_TO_VOXWARE (sound.volume_right) << 8));
				case SOUND_MIXER_READ_BASS:
					return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.bass));
				case SOUND_MIXER_READ_TREBLE:
					return IOCTL_OUT (arg, TONE_DB_TO_VOXWARE (sound.treble));
				case SOUND_MIXER_READ_SPEAKER: { int porta;
					if ((boot_info.bi_atari.mch_cookie >> 16) == ATARI_MCH_TT) {
						cli();
						sound_ym.rd_data_reg_sel = 14;
						porta = sound_ym.rd_data_reg_sel;
						sti();
						return IOCTL_OUT (arg, porta & 0x40 ? 0 : 100);
					}
					else
						return -EINVAL;
					}
				case SOUND_MIXER_WRITE_VOLUME:
					return IOCTL_OUT (arg, sound_set_volume (IOCTL_IN (arg)));
				case SOUND_MIXER_WRITE_BASS:
					return IOCTL_OUT (arg, sound_set_bass (IOCTL_IN (arg)));
				case SOUND_MIXER_WRITE_TREBLE:
					return IOCTL_OUT (arg, sound_set_treble (IOCTL_IN (arg)));
				case SOUND_MIXER_WRITE_SPEAKER:
					if ((boot_info.bi_atari.mch_cookie >> 16) == ATARI_MCH_TT) {
						int porta;
						cli();
						sound_ym.rd_data_reg_sel = 14;
						porta = (sound_ym.rd_data_reg_sel & ~0x40) |
					            (IOCTL_IN(arg) < 50 ? 0x40 : 0);
						sound_ym.wd_data = porta;
						sti();
						return IOCTL_OUT (arg, porta & 0x40 ? 0 : 100);
					} else
						return -EINVAL;
			}
			break;

		case DMASND_AMIGA:
	}

	return -EINVAL;
}



/*
 * sound queue stuff, the heart of the driver
 */


struct sound_buffer {
	char *start;
	int size;
};


struct sound_queue {
	int max_count, block_size;
	char *bufStart;

	/* it shouldn't be necessary to declare any of these volatile */
	int front, rear, count;
	int rear_size;
	int playing;
	struct wait_queue *write_queue, *open_queue, *sync_queue;
	int open_mode;
	int busy, syncing;
	int ignore_int;		/* ++TeSche: used for Falcon */
};


static struct sound_queue sq;


#define sq_block_address(i) ((char *)sq.bufStart + (i) * sq.block_size)
#define SIGNAL_RECEIVED (current->signal & ~current->blocked)
#define NON_BLOCKING(open_mode) (open_mode & O_NONBLOCK)
#define ONE_SECOND HZ	/* in jiffies (100ths of a second) */
#define NO_TIME_LIMIT 0xffffffff
#define SLEEP(queue, time_limit) \
	current->timeout = jiffies + (time_limit); \
	interruptible_sleep_on (&queue);
#define WAKE_UP(queue) (wake_up_interruptible (&queue))

#define	DISABLE_IRQ()							\
	do {								\
		mfp.int_mk_a &= ~0x20;					\
	} while( 0 )

#define	ENABLE_IRQ()							\
	do {								\
		mfp.int_mk_a |= 0x20;					\
	} while( 0 )


static void sq_init (int numBufs, int bufSize, char *bufStart)
{
	sq.max_count = numBufs;
	sq.block_size = bufSize;
	sq.bufStart = bufStart;

	sq.front = sq.count = 0;
	sq.rear = -1;
	sq.write_queue = sq.open_queue = sq.sync_queue = 0;
	sq.busy = 0;
	sq.syncing = 0;
	sq.ignore_int = 0;
	sq.playing = 0;

	sound_silence();

	/* whatever you like as startup mode for /dev/dsp,
	 * (/dev/audio hasn't got a startup mode). note that
	 * once changed a new open() will *not* restore these!
	 */
	sound.dsp.format = AFMT_S8;
	sound.dsp.stereo = 0;
	sound.dsp.size = 8;

	/* set minimum rate possible without expanding */
	if (sound.mach.type == DMASND_FALCON)
		sound.dsp.speed = 8195;
	else
		sound.dsp.speed = 6258;

	/* before the first open to /dev/dsp this wouldn't be set */
	sound.soft = sound.dsp;
}


static void sq_play_next_frame (int index)
{
	char *start, *end;

	/* used by sq_play() if all doubts whether there really is something
	 * to be played are already wiped out.
	 */
	start = sq_block_address (sq.front);
	end = start + ((sq.count == index)? sq.rear_size: sq.block_size);
	DMASNDSetEnd (VTOP (end));
	DMASNDSetBase (VTOP (start));
		/* Since only an even number of samples per frame can
		be played, we might lose one byte here. (TO DO) */
	sq.front = (sq.front + 1) % sq.max_count;
	sq.playing++;
	tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_REPEAT;
}


static void sq_play (void)
{
	/* ++TeSche: Note that sq.playing is no longer just a flag but holds
	 * the number of frames the DMA is currently programmed for instead,
	 * may be 0, 1 (currently being played) or 2 (pre-programmed).
	 *
	 * Chances done to sq.count and sq.playing are a bit more subtle again
	 * so now I must admit I also prefer disabling the irq here rather
	 * than considering all possible situations. But the point is that
	 * disabling the irq doesn't have any bad influence on this version of
	 * the driver as we benefit from having pre-programmed the DMA
	 * whereever possible: There's no need to reload the DMA at the exact
	 * time of an interrupt but only at some time while the pre-programmed
	 * frame in playing!
	 */
	DISABLE_IRQ();

	if (sq.playing == 2 ||	/* DMA is 'full' */
	    sq.count <= 0) {	/* nothing to do */
		ENABLE_IRQ();
		return;
	}

	if (sq.playing == 0) {
		/* looks like there's nothing 'in' the DMA yet, so try
		 * to put two frames into it (at least one is available).
		 */
		if (sq.count == 1 &&
		    sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, the only existing frame is not
			 * yet filled and we're not syncing?
			 */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(1);
		if (sq.count == 1) {
			/* no more frames */
			ENABLE_IRQ();
			return;
		}
		if (sq.count == 2 &&
		    sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, there were two frames, but the second
			 * one is not yet filled and we're not syncing?
			 */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(2);
	} else {
		/* there's already a frame being played so we may only stuff
		 * one new into the DMA, but even if this may be the last
		 * frame existing the previous one is still on sq.count.
		 */
		if (sq.count == 2 &&
		    sq.rear_size < sq.block_size &&
		    !sq.syncing) {
			/* hmmm, the only existing frame is not
			 * yet filled and we're not syncing?
			 */
			ENABLE_IRQ();
			return;
		}
		sq_play_next_frame(2);
	}
	ENABLE_IRQ();
}


static void sq_interrupt (void)
{
#if 0
	/* ++TeSche: if you should want to test this... */
	static int cnt = 0;
	if (sq.playing == 2)
		if (++cnt == 10) {
			/* simulate losing an interrupt */
			cnt = 0;
			return;
		}
#endif

	if (sq.ignore_int && (sound.mach.type == DMASND_FALCON)) {
		/* ++TeSche: Falcon only: ignore first irq because it comes
		 * immediately after starting a frame. after that, irqs come
		 * (almost) like on the TT.
		 */
		sq.ignore_int = 0;
		return;
	}

	if (!sq.playing) {
		/* playing was interrupted and sq_reset() has already cleared
		 * the sq variables, so better don't do anything here.
		 */
		WAKE_UP (sq.sync_queue);
		return;
	}

	/* Probably ;) one frame is finished. Well, in fact it may be that a
	 * pre-programmed one is also finished because there has been a long
	 * delay in interrupt delivery and we've completely lost one, but
	 * there's no way to detect such a situation. In such a case the last
	 * frame will be played more than once and the situation will recover
	 * as soon as the irq gets through.
	 */
	sq.count--;
	sq.playing--;

	if (!sq.playing) {
		tt_dmasnd.ctrl = DMASND_CTRL_OFF;
		sq.ignore_int = 1;
	}

	WAKE_UP (sq.write_queue);
		/* At least one block of the queue is free now
		so wake up a writing process blocked because
		of a full queue. */

	if ((sq.playing != 1) || (sq.count != 1))
		/* We must be a bit carefully here: sq.count indicates the
		 * number of buffers used and not the number of frames to
		 * be played. If sq.count==1 and sq.playing==1 that means
		 * the only remaining frame was already programmed earlier
		 * (and is currently running) so we mustn't call sq_play()
		 * here, otherwise we'll play one frame too much.
		 */
		sq_play();

	if (!sq.playing) WAKE_UP (sq.sync_queue);
		/* We are not playing after sq_play (), so there
		is nothing to play any more. Wake up a process
		waiting for audio output to drain. */
}


/* ++TeSche: radically changed this one too */

static int sq_write (const char *src, int uLeft)
{
	int uWritten = 0, bLeft;
	char *dest;
	long uUsed, bUsed;

	/* ++TeSche: Is something like this necessary?
	 * Hey, that's an honest question! Or does any other part of the
	 * filesystem already checks this situation? I really don't know.
	 */
	if (uLeft < 1)
		return 0;

	/* The interrupt doesn't start to play the last, incomplete frame.
	 * Thus we can append to it without disabling the interrupts! (Note
	 * also that sq.rear isn't affected by the interrupt.)
	 */

	if ((sq.count > 0) && ((bLeft = sq.block_size - sq.rear_size) > 0)) {

		dest = sq_block_address(sq.rear) + sq.rear_size;
		(*sound.copyTranslate)(uLeft, src, bLeft, dest, &uUsed, &bUsed);
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		sq.rear_size += bUsed;
	}

	do {
		if (sq.count == sq.max_count) {
			sq_play();
			if (NON_BLOCKING (sq.open_mode))
				return (uWritten > 0) ? uWritten : -EAGAIN;
			SLEEP (sq.write_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)
				return (uWritten > 0) ? uWritten : -EINTR;
		}

		/* Here, we can avoid disabling the interrupt by first
		 * copying and translating the data, and then updating
		 * the sq variables. Until this is done, the interrupt
		 * won't see the new frame and we can work on it
		 * undisturbed.
		 */

		bLeft = sq.block_size;
		dest = sq_block_address ((sq.rear + 1) % sq.max_count);
		(*sound.copyTranslate)(uLeft, src, bLeft, dest, &uUsed, &bUsed);
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		if (bUsed) {
			sq.rear = (sq.rear + 1) % sq.max_count;
			sq.rear_size = bUsed;
			sq.count++;
		}

	} while (bUsed);   /* uUsed may have been 0 */

	sq_play();

	return uWritten;
}


static int sq_open (int open_mode)
{
	if (sq.busy) {
		if (NON_BLOCKING (open_mode)) return -EBUSY;
		while (sq.busy) {
			SLEEP (sq.open_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)  return -EINTR;
		}
	}
	sq.open_mode = open_mode;
	sq.busy = 1;
	sq.ignore_int = 1;
	return 0;
}


static void sq_reset (void)
{
	sound_silence ();
	sq.playing = 0;
	sq.count = 0;
	sq.front = (sq.rear + 1) % sq.max_count;
}


static int sq_sync (void)
{
	int rc = 0;

	sq.syncing = 1;
	sq_play();	/* there may be an incomplete frame waiting */

	while (sq.playing) {
		SLEEP (sq.sync_queue, ONE_SECOND);
		if (SIGNAL_RECEIVED) {
			/* While waiting for audio output to drain,
			an interrupt occured. Stop audio output
			immediately and clear the queue. */
			sq_reset ();
			rc = -EINTR;
			break;
		}
	}

	sq.syncing = 0;
	return rc;
}


static int sq_release (void)
{
	int rc = 0;
	if (sq.busy) {
		rc = sq_sync ();
		sq.busy = 0;
		WAKE_UP (sq.open_queue);
			/* Wake up a process waiting for the queue beeing
			released. Note: There may be several processes
			waiting for a call to open () returning. */
	}
	return rc;
}



/*
 * /dev/sndstat
 */


struct sound_state {
	int busy;
	char buf [512];
	int len, ptr;
};


struct sound_state state;


static int state_puts (const char *s)
{
	int l;
	for (l = 0; l < 256 && s [l]; l++);
		/* l = max (256, strlen (s)) */
	if (state.len + l > sizeof (state.buf)) return 0;
	memcpy (&state.buf [state.len], s, l);
	state.len += l;
	return 1;
}


static int state_puti (int i, int radix)
{
	int l;
	static char digit [] = "0123456789abcdef";
	char buf [11];

	l = 0; buf [sizeof (buf) - 1] = 0;

	if (i < 0) {
		state_puts ("-");
		i = -i;
	}

	do {
		buf [sizeof (buf) - 2 - l] = digit [i % radix]; l++;
		i /= radix;
	} while (i);

	return state_puts (&buf [sizeof (buf) - 1 - l]);
}


static void state_init (void)
{
	state.busy = 0;
}


static int state_open (int open_mode)
{
	int	count, rear_size, playing, syncing;
	
	if (state.busy) return -EBUSY;

	state.len = 0;
	state.ptr = 0;
	state.busy = 1;

	state_puts ("Atari DMA sound driver:\n");

	state_puts ("\tsound.format = 0x");
	state_puti (sound.soft.format, 16); 
	switch (sound.soft.format) {
		case AFMT_MU_LAW:
			state_puts(" (mu-law)");
			break;
		case AFMT_U8:
			state_puts(" (unsigned 8 bit)");
			break;
		case AFMT_S8:
			state_puts(" (signed 8 bit)");
			break;
		case AFMT_S16_BE:
			state_puts(" (signed 16 bit big)");
			break;
		case AFMT_U16_BE:
			state_puts(" (unsigned 16 bit big)");
			break;
	}
	state_puts ("\n");
	
	state_puts ("\tsound.speed = ");
	state_puti (sound.soft.speed, 10);
	state_puts ("Hz (phys. ");
	state_puti (sound.hard.speed, 10);
	state_puts ("Hz)\n");

	state_puts ("\tsound.stereo = 0x");
	state_puti (sound.soft.stereo, 16);
	state_puts (sound.soft.stereo ? " (stereo)\n" : " (mono)\n");

	switch (sound.mach.type) {
		case DMASND_TT:
			state_puts ("\tsound.volume_left = ");
			state_puti (sound.volume_left, 10); state_puts (" db [-40...0]\n");
			state_puts ("\tsound.volume_right = ");
			state_puti (sound.volume_right, 10); state_puts (" db [-40...0]\n");
			state_puts ("\tsound.bass = ");
			state_puti (sound.bass, 10); state_puts (" db [-12...+12]\n");
			state_puts ("\tsound.treble = ");
			state_puti (sound.treble, 10); state_puts (" db [-12...+12]\n");
			break;
		case DMASND_FALCON:
			state_puts ("\tsound.volume_left = ");
			state_puti (sound.volume_left, 10); state_puts (" dB [-22.5...0]\n");
		  	state_puts ("\tsound.volume_right = ");
			state_puti (sound.volume_right, 10); state_puts (" dB [-22.5...0]\n");
			break;
		case DMASND_AMIGA:
	}

	state_puts( "\tsq.block_size = ");
	state_puti( sq.block_size, 10 );
	state_puts( " sq.max_count = ");
	state_puti( sq.max_count, 10 );
	state_puts( "\n" );

	count     = sq.count;
	rear_size = sq.rear_size;
	playing   = sq.playing;
	syncing   = sq.syncing;

	state_puts( "\tsq.count = ");
	state_puti( count, 10 );
	state_puts( " sq.rear_size = ");
	state_puti( rear_size, 10 );
	state_puts( "\n" );

	state_puts( "\tsq.playing = ");
	state_puti( playing, 10 );
	state_puts( " sq.syncing = ");
	state_puti( syncing, 10 );
	state_puts( "\n" );

	return 0;
}


static int state_release (void)
{
	state.busy = 0;
	return 0;
}


static int state_read (char *dest, int count)
{
	int n = state.len - state.ptr;
	if (n > count) n = count;
	if (n <= 0) return 0;
	memcpy_tofs (dest, &state.buf [state.ptr], n);
	state.ptr += n;
	return n;
}



/*
 * high level sound stuff
 */


static int sound_open (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev) & 0x0f;

	switch (dev) {
	case SND_DEV_STATUS: return state_open (file->f_flags);
	case SND_DEV_CTL: return mixer_open (file->f_flags);
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: {
		int rc = sq_open (file->f_flags);
		if (rc == 0) {
			sound.minDev = dev;
			sound.soft = sound.dsp;
			sound_init ();
			if (dev == SND_DEV_AUDIO) {
				sound_set_speed(8000);
				sound_set_stereo(0);
				sound_set_format(AFMT_MU_LAW);
			}
		}
		return rc;
	}
	default: return -ENXIO;
	}

}


static int sound_fsync(struct inode *inode, struct file *filp)
{
	int dev = MINOR (inode->i_rdev) & 0x0f;

	switch (dev) {
		case SND_DEV_STATUS:
		case SND_DEV_CTL:
			return 0;
		case SND_DEV_DSP:
		case SND_DEV_AUDIO:
			return sq_sync();
		default:
			printk ("sound_fsync: Unknown minor device %d\n", dev);
			return -EINVAL;
	}
}


static void sound_release (struct inode *inode, struct file *file)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: state_release (); return;
	case SND_DEV_CTL: mixer_release (); return;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO:
		sq_release (); sound.soft = sound.dsp; sound_silence();
		return;
	default:
		printk ("sound_release: Unknown minor device %d\n", dev);
	}
}


static int sound_lseek (struct inode *inode,
			struct file *file,
			off_t offset,
			int orig)
{
	return -EPERM;
}


static int sound_read (struct inode *inode,
		       struct file *file,
		       char *buf,
		       int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS: return state_read (buf, count);
	case SND_DEV_CTL:
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return -EPERM;
	default:
		printk ("sound_read: Unknown minor device %d\n", dev);
		return -EINVAL;
	}
}


static int sound_write (struct inode *inode,
			struct file *file,
			char *buf,
			int count)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {
	case SND_DEV_STATUS:
	case SND_DEV_CTL: return -EPERM;
	case SND_DEV_DSP:
	case SND_DEV_AUDIO: return sq_write(buf, count);
	default:
		printk ("sound_write: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

}


static int sound_select (struct inode *inode,
			 struct file *file,
			 int sel_type,
			 select_table *wait)
{
	return -EPERM;
}


static int ioctl_return (int *addr, int value)
{
	int error;

	if (value < 0) return value;

	error = verify_area (VERIFY_WRITE, addr, sizeof (int));
	if (error) return error;

	put_fs_long (value, (long*)&((addr)[0]));
	return 0;
}


static int sound_ioctl (struct inode *inode,
			struct file *file,
			unsigned int cmd,
			unsigned long arg)
{
	int dev = MINOR (inode->i_rdev);

	switch (dev & 0x0f) {

	case SND_DEV_STATUS: return -EPERM;
	case SND_DEV_CTL: return mixer_ioctl (inode, file, cmd, arg);

	case SND_DEV_AUDIO:
	case SND_DEV_DSP:
		switch (cmd) {

		case SNDCTL_DSP_RESET:
			sq_reset ();
			return 0;

		case SNDCTL_DSP_POST:
		case SNDCTL_DSP_SYNC:
			return sound_fsync(inode, file);

		/* ++TeSche: before changing any of these it's probably wise to
		 * wait until sound playing has settled down
		 */
		case SNDCTL_DSP_SPEED:
			sound_fsync(inode, file);
			return IOCTL_OUT (arg, sound_set_speed (IOCTL_IN (arg)));
		case SNDCTL_DSP_STEREO:
			sound_fsync(inode, file);
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg)));
		case SOUND_PCM_WRITE_CHANNELS:
			sound_fsync(inode, file);
			return IOCTL_OUT (arg, sound_set_stereo (IOCTL_IN (arg) - 1) + 1);
		case SNDCTL_DSP_SETFMT:
			sound_fsync(inode, file);
			return IOCTL_OUT (arg, sound_set_format (IOCTL_IN (arg)));

		case SNDCTL_DSP_GETFMTS:
			switch (sound.mach.type) {
				case DMASND_TT:
					return IOCTL_OUT (arg, AFMT_MU_LAW | AFMT_U8 | AFMT_S8);
				case DMASND_FALCON:
					return IOCTL_OUT (arg, AFMT_MU_LAW | AFMT_U8 | AFMT_S8 |
					                       AFMT_S16_BE | AFMT_U16_BE);
				case DMASND_AMIGA:
					return -ENXIO;
			}
		case SNDCTL_DSP_GETBLKSIZE:
			return IOCTL_OUT (arg, 10240);
		case SNDCTL_DSP_SUBDIVIDE:
		case SNDCTL_DSP_SETFRAGMENT: break;

		default: return mixer_ioctl (inode, file, cmd, arg);
		}
		break;

	default:
		printk ("sound_ioctl: Unknown minor device %d\n", dev);
		return -EINVAL;
	}

	return( -EINVAL );
}


static struct file_operations sound_fops =
{
	sound_lseek,
	sound_read,
	sound_write,
	NULL,
	sound_select,
	sound_ioctl,
	NULL,
	sound_open,
	sound_release,
	sound_fsync
};



/*** config & setup **********************************************************/


long soundcard_init (long mem_start)
{
	char *bufStart;

	if (ATARIHW_PRESENT (CODEC))
		sound.mach = machFalcon;
	else
		sound.mach = machTT;

	if (ATARIHW_PRESENT (PCM_8BIT) && (ATARIHW_PRESENT (MICROWIRE) || (sound.mach.type == DMASND_FALCON))) {
		if ((mfp.int_en_a & mfp.int_mk_a & 0x20) == 0) {

			/* Set up sound queue, /dev/audio and /dev/dsp.
			TO DO: Allocate buffer space dynamically.
			But I do not see how to do this, since kmalloc will not
			allocate more than 4K and vmalloc does not guarantee
			memory to be physically contiguous, which is necessary
			for DMA transfer. */
			/* ++roman: It would be veeery difficult to allocate
			 * these buffers dynamically... Need not only contigous
			 * pages, but also in ST-Ram...
			 */
			if (!(bufStart = (char *)atari_stram_alloc (numBufs * (bufSize << 10), &mem_start))) {
				printk("Atari DMA sound driver: Not enough buffer memory, driver disabled!\n");
				return mem_start;
			}

			/* Register driver with the VFS. */
			register_chrdev (SOUND_MAJOR, "sound", &sound_fops);
			printk ("Atari DMA sound driver installed, using %d buffers of %dk.\n",
				numBufs, bufSize);

			sq_init (numBufs, bufSize << 10, bufStart);
        
			/* Set up /dev/sndstat. */
			state_init ();

			/* Set up /dev/mixer. */
			mixer_init ();

			/* Set up timer A. Timer A
			will receive a signal upon end of playing from the sound
			hardware. Furthermore Timer A is able to count events
			and will cause an interrupt after a programmed number
			of events. So all we need to keep the music playing is
			to provide the sound hardware with new data upon
			an interrupt from timer A. */
			mfp.tim_ct_a = 0;	/* ++roman: Stop timer before programming! */
			mfp.tim_dt_a = 1;	/* Cause interrupt after first event. */
			mfp.tim_ct_a = 8;	/* Turn on event counting. */
			mach_add_isr (IRQ_MFP_TIMA, (isrfunc)sq_interrupt, 0, NULL,
						  "DMA sound"); /* Register interrupt handler. */
			mfp.int_en_a |= 0x20;	/* Turn interrupt on. */
			mfp.int_mk_a |= 0x20;
		} else
			printk ("Atari DMA sound driver: Timer A interrupt already in use\n");
	}

	return mem_start;
}


void atari_sound_setup (char *str, int *ints)
{
	/* check the bootstrap parameter for "atasound=" */

	switch (ints[0]) {
		case 3:
			if ((ints[3] < 0) || (ints[3] > MAX_CATCH_RADIUS)) {
				printk("atari_sound_setup: illegal catch radius, using default = %d\n", catchRadius);
			} else {
				catchRadius = ints[3];
			}
			/* fall through */
		case 2:
			if (ints[1] < MIN_BUFFERS) {
				printk("atari_sound_setup: illegal number of buffers, using default = %d\n", numBufs);
			} else {
				numBufs = ints[1];
			}
			if (ints[2] < MIN_BUFSIZE) {
				printk("atari_sound_setup: illegal buffer size, using default = %d\n", bufSize);
			} else {
				bufSize = ints[2];
			}
			break;
		case 0:
			break;
		default:
			printk("atari_sound_setup: illegal number of arguments\n");
	}
}

#endif /* CONFIG_ATARI_DMA_SOUND */
